import shutil
import subprocess  # nosec
from pathlib import Path
import json

import pandas as pd
import pyarrow as pa
from pyarrow import feather
import pytest
from ert_shared.plugins.plugin_manager import ErtPluginManager

import webviz_subsurface.ert_jobs.jobs


def _create_minimal_ert_config_file(tmp_path: Path, forward_model_string: str) -> Path:
    runpath = tmp_path / Path("output")
    ert_config_file = tmp_path / "config.ert"

    ert_config_file.write_text(
        f"""
ECLBASE          something
RUNPATH          {runpath}
NUM_REALIZATIONS 1
QUEUE_OPTION     LSF MAX_RUNNING 1
QUEUE_SYSTEM     LOCAL
FORWARD_MODEL    {forward_model_string}
"""
    )

    return ert_config_file


def test_export_connection_status(testdata_folder: Path, tmp_path: Path) -> None:

    eclbase = (
        testdata_folder
        / "01_drogon_ahm"
        / "realization-0"
        / "iter-0"
        / "eclipse"
        / "model"
        / "DROGON-0"
    ).resolve()
    assert eclbase.with_suffix(".UNSMRY").exists()

    ert_config_file = _create_minimal_ert_config_file(
        tmp_path,
        f"WELL_CONNECTION_STATUS(<ECLBASE>={eclbase})",
    )

    subprocess.check_output(["ert", "test_run", ert_config_file], cwd=tmp_path)  # nosec
    output_file = (
        tmp_path
        / "output"
        / "share"
        / "results"
        / "tables"
        / "well_connection_status.parquet"
    )
    assert output_file.exists()

    df = pd.read_parquet(output_file)
    assert list(df.columns) == ["DATE", "WELL", "I", "J", "K", "OP/SH"]


def test_smry2arrow(testdata_folder: Path, tmp_path: Path) -> None:

    eclbase = (
        testdata_folder
        / "01_drogon_ahm"
        / "realization-0"
        / "iter-0"
        / "eclipse"
        / "model"
        / "DROGON-0"
    ).resolve()
    assert eclbase.with_suffix(".UNSMRY").exists()

    output_file = tmp_path / "output.arrow"

    ert_config_file = _create_minimal_ert_config_file(
        tmp_path, f"SMRY2ARROW(<ECLBASE>={eclbase})"
    )
    output_file = tmp_path / "output" / "share" / "results" / "tables" / "unsmry.arrow"
    subprocess.check_output(["ert", "test_run", ert_config_file], cwd=tmp_path)  # nosec

    assert output_file.exists()

    table = feather.read_table(output_file)
    assert table.shape == (243, 921)

    sample_date = table["DATE"][0]
    assert sample_date.type == pa.timestamp("ms")

    schema = table.schema
    field = schema.field("FOPT")
    field_meta = json.loads(field.metadata[b"smry_meta"])
    assert field.type == pa.float32()
    assert field_meta["unit"] == "SM3"
    assert field_meta["is_total"] is True
    assert field_meta["is_rate"] is False
    assert field_meta["is_historical"] is False

    field = schema.field("FOPR")
    field_meta = json.loads(field.metadata[b"smry_meta"])
    assert field.type == pa.float32()
    assert field_meta["unit"] == "SM3/DAY"
    assert field_meta["is_total"] is False
    assert field_meta["is_rate"] is True
    assert field_meta["is_historical"] is False

    field = schema.field("FOPTH")
    field_meta = json.loads(field.metadata[b"smry_meta"])
    assert field.type == pa.float32()
    assert field_meta["unit"] == "SM3"
    assert field_meta["is_total"] is True
    assert field_meta["is_rate"] is False
    assert field_meta["is_historical"] is True


@pytest.fixture(name="expected_jobs")
def expected_jobs_fixture():
    """dictionary of installed jobs with location to config"""
    config_location = (
        Path(__file__).absolute().parent.parent.parent
        / "webviz_subsurface"
        / "ert_jobs"
    )
    expected_job_names = ["WELL_CONNECTION_STATUS", "SMRY2ARROW"]
    return {
        name: str(config_location / "config_jobs" / name) for name in expected_job_names
    }


# Avoid category inflation. Add to this list when it makes sense:
ACCEPTED_JOB_CATEGORIES = ["modelling", "utility"]


def test_hook_implementations(expected_jobs):
    """Test that we have the correct set of jobs installed,
    nothing more, nothing less"""
    plugin_m = ErtPluginManager(plugins=[webviz_subsurface.ert_jobs.jobs])

    installable_jobs = plugin_m.get_installable_jobs()
    for wf_name, wf_location in expected_jobs.items():
        assert wf_name in installable_jobs
        assert installable_jobs[wf_name].endswith(wf_location)
        assert Path(installable_jobs[wf_name]).exists()

    assert set(installable_jobs.keys()) == set(expected_jobs.keys())

    expected_workflow_jobs = {}
    installable_workflow_jobs = plugin_m.get_installable_workflow_jobs()
    for wf_name, wf_location in expected_workflow_jobs.items():
        assert wf_name in installable_workflow_jobs
        assert installable_workflow_jobs[wf_name].endswith(wf_location)

    assert set(installable_workflow_jobs.keys()) == set(expected_workflow_jobs.keys())


def test_job_config_syntax(expected_jobs):
    """Check for syntax errors made in job configuration files"""
    for job_config in expected_jobs.values():
        # Check (loosely) that double-dashes are enclosed in quotes:
        for line in Path(job_config).read_text().splitlines():
            if not line.strip().startswith("--") and "--" in line:
                assert '"--' in line and " --" not in line


def test_executables(expected_jobs):
    """Test executables listed in job configurations exist in $PATH"""
    for job_config in expected_jobs.values():
        executable = Path(job_config).read_text().splitlines()[0].split()[1]
        assert shutil.which(executable)


def test_hook_implementations_job_docs():
    """For each installed job, we require the associated
    description string to be nonempty, and valid RST markup"""

    plugin_m = ErtPluginManager(plugins=[webviz_subsurface.ert_jobs.jobs])

    installable_jobs = plugin_m.get_installable_jobs()

    docs = plugin_m.get_documentation_for_jobs()

    assert set(docs.keys()) == set(installable_jobs.keys())

    for job_name in installable_jobs.keys():
        desc = docs[job_name]["description"]
        assert desc != ""
        category = docs[job_name]["category"]
        assert category != "other"
        assert category.split(".")[0] in ACCEPTED_JOB_CATEGORIES



from dash import html
import webviz_core_components as wcc
from webviz_subsurface._models import InplaceVolumesModel


def tornado_selections_layout(
    uuid: str, volumemodel: InplaceVolumesModel, tab: str
) -> html.Div:
    """Layout for selecting tornado data"""
    return html.Div(
        children=[
            tornado_controls_layout(uuid, tab, volumemodel),
            settings_layout(uuid, tab, volumemodel),
        ]
    )


def tornado_controls_layout(
    uuid: str, tab: str, volumemodel: InplaceVolumesModel
) -> wcc.Selectors:
    mode_options = [{"label": "Custom", "value": "custom"}]
    if "BULK" in volumemodel.responses:
        mode_options.append({"label": "Bulk vs STOIIP/GIIP", "value": "locked"})
    return wcc.Selectors(
        label="TORNADO CONTROLS",
        children=[
            wcc.RadioItems(
                label="Mode:",
                id={"id": uuid, "tab": tab, "selector": "mode"},
                options=mode_options,
                value="custom",
            ),
            tornado_selection(
                position="left", uuid=uuid, tab=tab, volumemodel=volumemodel
            ),
            tornado_selection(
                position="right", uuid=uuid, tab=tab, volumemodel=volumemodel
            ),
        ],
    )


def tornado_selection(
    position: str, uuid: str, tab: str, volumemodel: InplaceVolumesModel
) -> html.Div:
    return html.Div(
        style={"margin-top": "10px"},
        children=[
            html.Label(
                children=f"Tornado ({position})",
                style={"display": "block"},
                className="webviz-underlined-label",
            ),
            wcc.Dropdown(
                id={"id": uuid, "tab": tab, "selector": f"Response {position}"},
                clearable=False,
                value=volumemodel.responses[0],
            ),
            html.Details(
                open=False,
                children=[
                    html.Summary("Sensitivity filter"),
                    wcc.Select(
                        id={
                            "id": uuid,
                            "tab": tab,
                            "selector": f"Sensitivities {position}",
                        },
                        options=[
                            {"label": i, "value": i} for i in volumemodel.sensitivities
                        ],
                        value=volumemodel.sensitivities,
                        size=min(15, len(volumemodel.sensitivities)),
                    ),
                ],
            ),
        ],
    )


def settings_layout(
    uuid: str, tab: str, volumemodel: InplaceVolumesModel
) -> wcc.Selectors:
    return wcc.Selectors(
        label="⚙️ SETTINGS",
        open_details=True,
        children=[
            wcc.Dropdown(
                label="Scale:",
                id={"id": uuid, "tab": tab, "selector": "Scale"},
                options=[
                    {"label": "Relative value (%)", "value": "Percentage"},
                    {"label": "Relative value", "value": "Absolute"},
                    {"label": "True value", "value": "True"},
                ],
                value="Percentage",
                clearable=False,
            ),
            html.Div(
                style={"margin-top": "10px"},
                children=wcc.Checklist(
                    id={"id": uuid, "tab": tab, "selector": "real_scatter"},
                    options=[{"label": "Show realization points", "value": "Show"}],
                    value=[],
                ),
            ),
            cut_by_ref(uuid, tab),
            labels_display(uuid, tab),
            wcc.Dropdown(
                label="Reference:",
                id={"id": uuid, "tab": tab, "selector": "Reference"},
                options=[
                    {"label": elm, "value": elm} for elm in volumemodel.sensitivities
                ],
                value="rms_seed"
                if "rms_seed" in volumemodel.sensitivities
                else volumemodel.sensitivities[0],
                clearable=False,
            ),
        ],
    )


def cut_by_ref(uuid: str, tab: str) -> html.Div:
    return html.Div(
        style={"margin-bottom": "10px"},
        children=wcc.Checklist(
            id={"id": uuid, "tab": tab, "selector": "Remove no impact"},
            options=[
                {"label": "Remove sensitivities with no impact", "value": "Remove"}
            ],
            value=["Remove"],
        ),
    )


def labels_display(uuid: str, tab: str) -> html.Div:
    return html.Div(
        style={"margin-bottom": "10px"},
        children=[
            wcc.RadioItems(
                label="Label options:",
                id={"id": uuid, "tab": tab, "selector": "labeloptions"},
                options=[
                    {"label": "detailed", "value": "detailed"},
                    {"label": "simple", "value": "simple"},
                    {"label": "hide", "value": "hide"},
                ],
                vertical=False,
                value="detailed",
            ),
        ],
    )



from typing import List
import dash


def create_range_string(real_list: list) -> str:
    idx = 0
    ranges = [[real_list[0], real_list[0]]]
    for real in list(real_list):
        if ranges[idx][1] in (real, real - 1):
            ranges[idx][1] = real
        else:
            ranges.append([real, real])
            idx += 1

    return ", ".join(
        map(lambda p: f"{p[0]}-{p[1]}" if p[0] != p[1] else str(p[0]), ranges)
    )


def update_relevant_components(id_list: list, update_info: List[dict]) -> list:
    output_id_list = [dash.no_update] * len(id_list)
    for elm in update_info:
        for idx, x in enumerate(id_list):
            if all(x[key] == value for key, value in elm["conditions"].items()):
                output_id_list[idx] = elm["new_value"]
                break
    return output_id_list


def move_to_end_of_list(element: str, list_of_elements: list) -> list:
    return [x for x in list_of_elements if x != element] + [element]



from typing import List, Optional

from dash import html
import dash_bootstrap_components as dbc


def modal_layout(
    uuid: str,
    modal_id: str,
    title: str,
    body_children: List,
    footer: Optional[dbc.ModalFooter] = None,
    size: str = "sm",
) -> dbc.Modal:
    modalchildren = [
        dbc.ModalHeader(title),
        dbc.ModalBody(
            children=[*body_children],
        ),
    ]
    if footer:
        modalchildren.append(footer)
    return dbc.Modal(
        style={"marginTop": "20vh"},
        children=modalchildren,
        id={"id": uuid, "modal_id": modal_id, "element": "wrapper"},
        size=size,
    )


def open_modal_layout(uuid: str, modal_id: str, title: str) -> dbc.Button:
    return html.Div(
        children=html.Button(
            title,
            className="webviz-structunc-open-modal-btn",
            id={"id": uuid, "modal_id": modal_id, "element": "button-open"},
        ),
    )


def clear_all_apply_modal_buttons(
    uuid: str, modal_id: str, apply_disabled: bool = True
) -> dbc.ModalFooter:
    return dbc.ModalFooter(
        children=[
            html.Div(
                children=[
                    dbc.Button(
                        "Clear",
                        style={"padding": "0 20px"},
                        className="mr-1",
                        id={"id": uuid, "modal_id": modal_id, "element": "clear"},
                    ),
                    dbc.Button(
                        "All",
                        style={"padding": "0 20px"},
                        className="mr-1",
                        id={"id": uuid, "modal_id": modal_id, "element": "all"},
                    ),
                    dbc.Button(
                        "Apply",
                        style={"padding": "0 20px", "visibility": "hidden"}
                        if apply_disabled
                        else {"padding": "0 20px"},
                        className="mr-1",
                        id={"id": uuid, "modal_id": modal_id, "element": "apply"},
                    ),
                ]
            ),
        ]
    )



from typing import Dict, Optional
from pathlib import Path
import os
import hashlib
import json
import logging
import pickle  # nosec
from enum import Enum

import pandas as pd
from fmu.ensemble import ScratchEnsemble

from webviz_config.webviz_factory_registry import WEBVIZ_FACTORY_REGISTRY
from webviz_config.webviz_factory import WebvizFactory
from webviz_config.webviz_instance_info import WebvizRunMode

from .ensemble_table_provider import EnsembleTableProvider
from .ensemble_table_provider import EnsembleTableProviderSet
from .ensemble_table_provider_impl_arrow import EnsembleTableProviderImplArrow
from .ensemble_table_provider_impl_inmem_parquet import (
    EnsembleTableProviderImplInMemParquet,
)
from .._utils.perf_timer import PerfTimer


class BackingType(Enum):
    ARROW = "arrow"
    INMEM_PARQUET = "inmem_parquet"


LOGGER = logging.getLogger(__name__)


def _make_hash_string(string_to_hash: str) -> str:
    # There is no security risk here and chances of collision should be very slim
    return hashlib.md5(string_to_hash.encode()).hexdigest()  # nosec


class EnsembleTableProviderFactory(WebvizFactory):
    def __init__(
        self,
        root_storage_folder: Path,
        backing_type: BackingType,
        allow_storage_writes: bool,
    ) -> None:

        self._storage_dir = Path(root_storage_folder) / __name__
        self._backing_type: BackingType = backing_type
        self._allow_storage_writes = allow_storage_writes
        self._scratch_ensemble_cache: Dict[str, bytes] = {}

        LOGGER.info(
            f"EnsembleTableProviderFactory init: backing_type={repr(self._backing_type)}"
        )
        LOGGER.info(
            f"EnsembleTableProviderFactory init: storage_dir={self._storage_dir}"
        )

        if self._allow_storage_writes:
            # For now, just make sure the storage folder exists
            os.makedirs(self._storage_dir, exist_ok=True)

    @staticmethod
    def instance() -> "EnsembleTableProviderFactory":
        factory = WEBVIZ_FACTORY_REGISTRY.get_factory(EnsembleTableProviderFactory)

        if not factory:
            app_instance_info = WEBVIZ_FACTORY_REGISTRY.app_instance_info
            storage_folder = app_instance_info.storage_folder
            backing_type = BackingType.ARROW
            allow_writes = app_instance_info.run_mode != WebvizRunMode.PORTABLE

            my_settings = WEBVIZ_FACTORY_REGISTRY.all_factory_settings.get(
                "EnsembleTableProviderFactory"
            )
            if my_settings:
                LOGGER.info(
                    f"Parsing settings for EnsembleTableProviderFactory: {my_settings}"
                )
                if "backing_type" in my_settings:
                    backing_type = BackingType(my_settings["backing_type"])

            factory = EnsembleTableProviderFactory(
                storage_folder, backing_type, allow_writes
            )
            WEBVIZ_FACTORY_REGISTRY.set_factory(EnsembleTableProviderFactory, factory)

        return factory

    def cleanup_resources_after_plugin_init(self) -> None:
        self._scratch_ensemble_cache.clear()

    def create_provider_set_from_aggregated_csv_file(
        self,
        aggr_csv_file: Path,
    ) -> EnsembleTableProviderSet:

        LOGGER.info(f"create_provider_set_from_aggregated_csv_file() - {aggr_csv_file}")

        hashval = _make_hash_string(str(aggr_csv_file))
        main_storage_key = f"aggr_csv__{hashval}"

        storage_keys_to_load: Dict[str, str] = {}
        json_fn = self._storage_dir / (main_storage_key + ".json")
        try:
            with open(json_fn, "r") as file:
                storage_keys_to_load = json.load(file)
        except FileNotFoundError:
            # We can only recover from this if we're allowed to write to storage
            if not self._allow_storage_writes:
                raise

        if not storage_keys_to_load and self._allow_storage_writes:
            aggregated_df = pd.read_csv(aggr_csv_file)
            ensemble_names = aggregated_df["ENSEMBLE"].unique()

            LOGGER.info(
                f"Saving {len(ensemble_names)} table providers from aggregated CSV to backing store"
            )

            for ens_name in ensemble_names:
                storage_key = main_storage_key + "__" + ens_name
                ensemble_df = aggregated_df[aggregated_df["ENSEMBLE"] == ens_name]
                self._write_data_to_backing_store(storage_key, ensemble_df)
                storage_keys_to_load[ens_name] = storage_key

            with open(json_fn, "w") as file:
                json.dump(storage_keys_to_load, file)

        created_providers: Dict[str, EnsembleTableProvider] = {}
        for ens_name, storage_key in storage_keys_to_load.items():
            provider = self._create_provider_instance_from_backing_store(storage_key)
            if provider:
                created_providers[ens_name] = provider

        num_missing_models = len(storage_keys_to_load) - len(created_providers)
        if num_missing_models > 0:
            raise ValueError(f"Failed to load data for {num_missing_models} ensembles")

        LOGGER.info(f"Loaded {len(created_providers)} providers from backing store")

        return EnsembleTableProviderSet(created_providers)

    def create_provider_set_from_per_realization_csv_file(
        self, ensembles: Dict[str, str], csv_file_rel_path: str
    ) -> EnsembleTableProviderSet:
        # pylint: disable=too-many-locals

        LOGGER.info(
            f"create_provider_set_from_per_realization_csv_file() - {csv_file_rel_path}"
        )
        timer = PerfTimer()

        # Try and create/load providers from backing store
        created_providers: Dict[str, EnsembleTableProvider] = {}
        missing_storage_keys: Dict[str, str] = {}
        for ens_name, ens_path in ensembles.items():
            hashval = _make_hash_string(ens_path + csv_file_rel_path)
            storage_key = f"ens_csv__{hashval}"
            provider = self._create_provider_instance_from_backing_store(storage_key)
            if provider:
                created_providers[ens_name] = provider
                LOGGER.info(f"Loaded table provider for {ens_name} from backing store")
            else:
                missing_storage_keys[ens_name] = storage_key

        # If there are remaining keys AND we're allowed to write to storage,
        # we'll load the csv, write data to storage and then try and load again
        if missing_storage_keys and self._allow_storage_writes:
            for ens_name, storage_key in dict(missing_storage_keys).items():
                timer.lap_s()
                ens_path = ensembles[ens_name]
                scratch_ensemble = self._get_or_create_scratch_ensemble(
                    ens_name, ens_path
                )
                elapsed_create_scratch_ens_s = timer.lap_s()

                ensemble_df = scratch_ensemble.load_csv(csv_file_rel_path)
                del scratch_ensemble
                elapsed_load_csv_s = timer.lap_s()

                self._write_data_to_backing_store(storage_key, ensemble_df)
                provider = self._create_provider_instance_from_backing_store(
                    storage_key
                )
                elapsed_write_s = timer.lap_s()

                if provider:
                    created_providers[ens_name] = provider
                    del missing_storage_keys[ens_name]
                    LOGGER.info(
                        f"Saved table provider for {ens_name} to backing store ("
                        f"create_scratch_ens={elapsed_create_scratch_ens_s:.2f}s "
                        f"load_csv={elapsed_load_csv_s:.2f}s "
                        f"write={elapsed_write_s:.2f}s)"
                    )

        if missing_storage_keys:
            raise ValueError(
                f"Failed to load/create provider(s) for {len(missing_storage_keys)} ensembles"
            )

        LOGGER.info(
            f"create_provider_set_from_per_realization_csv_file() "
            f"- total time: {timer.elapsed_s():.2f}s"
        )

        return EnsembleTableProviderSet(created_providers)

    # Function to load the parameters.txt/json file using fmu.ensemble.
    # Should this be a separate Factory class? Typically both csv files and
    # the parameter file will be loaded from the same ensemble. Currently the ensembles
    # will be loaded and deleted twice if that is the case...
    def create_provider_set_from_per_realization_parameter_file(
        self, ensembles: Dict[str, str]
    ) -> EnsembleTableProviderSet:

        LOGGER.info("create_provider_set_from_per_realization_parameter_file() ...")

        timer = PerfTimer()

        storage_keys_to_load: Dict[str, str] = {}
        for ens_name, ens_path in ensembles.items():
            hashval = _make_hash_string(ens_path + "parameters")
            storage_keys_to_load[ens_name] = f"parameters__{hashval}"

        # We'll add all the models for the model set to this dictionary as we go
        created_providers: Dict[str, EnsembleTableProvider] = {}

        # First, try and load models from backing store
        for ens_name, storage_key in dict(storage_keys_to_load).items():
            provider = self._create_provider_instance_from_backing_store(storage_key)
            if provider:
                created_providers[ens_name] = provider
                del storage_keys_to_load[ens_name]
                LOGGER.info(f"Loaded table provider for {ens_name} from backing store")

        # If there are remaining keys to load and we're allowed to write to storage,
        # we'll load the parameters file, write data to storage and then try and load again
        if storage_keys_to_load and self._allow_storage_writes:
            for ens_name, storage_key in dict(storage_keys_to_load).items():
                timer.lap_s()
                ens_path = ensembles[ens_name]
                scratch_ensemble = self._get_or_create_scratch_ensemble(
                    ens_name, ens_path
                )
                elapsed_create_scratch_ens_s = timer.lap_s()

                ensemble_df = scratch_ensemble.parameters
                del scratch_ensemble
                elapsed_load_parameters_s = timer.lap_s()

                self._write_data_to_backing_store(storage_key, ensemble_df)
                provider = self._create_provider_instance_from_backing_store(
                    storage_key
                )
                elapsed_write_s = timer.lap_s()

                if provider:
                    created_providers[ens_name] = provider
                    del storage_keys_to_load[ens_name]
                    LOGGER.info(
                        f"Saved table provider for {ens_name} to backing store ("
                        f"create_scratch_ens={elapsed_create_scratch_ens_s:.2f}s "
                        f"load_parameters={elapsed_load_parameters_s:.2f}s "
                        f"write={elapsed_write_s:.2f}s)"
                    )

        # Should not be any remaining keys
        if storage_keys_to_load:
            raise ValueError(
                f"Failed to load data for {len(storage_keys_to_load)} ensembles"
            )

        LOGGER.info(
            f"create_provider_set_from_per_realization_parameter_file() "
            f"- total time: {timer.elapsed_s():.2f}s"
        )

        return EnsembleTableProviderSet(created_providers)

    def _create_provider_instance_from_backing_store(
        self, storage_key: str
    ) -> Optional[EnsembleTableProvider]:
        if self._backing_type == BackingType.ARROW:
            return EnsembleTableProviderImplArrow.from_backing_store(
                self._storage_dir, storage_key
            )
        if self._backing_type == BackingType.INMEM_PARQUET:
            return EnsembleTableProviderImplInMemParquet.from_backing_store(
                self._storage_dir, storage_key
            )

        raise ValueError("Unhandled backing type")

    def _write_data_to_backing_store(
        self, storage_key: str, ensemble_df: pd.DataFrame
    ) -> None:
        if self._backing_type == BackingType.ARROW:
            EnsembleTableProviderImplArrow.write_backing_store_from_ensemble_dataframe(
                self._storage_dir, storage_key, ensemble_df
            )
        elif self._backing_type == BackingType.INMEM_PARQUET:
            EnsembleTableProviderImplInMemParquet.write_backing_store_from_ensemble_dataframe(
                self._storage_dir, storage_key, ensemble_df
            )
        else:
            raise ValueError("Unhandled backing type")

    def _get_or_create_scratch_ensemble(
        self, ens_name: str, ens_path: str
    ) -> ScratchEnsemble:
        """Either creates a new ScratchEnsemble or retrieves a previously created one
        from our internal cache. Note that we cache a pickled version of ScratchEnsembles
        to avoid memory bloating as data is loaded and internalized by the ensemble
        """
        key = json.dumps({"ens_name": ens_name, "ens_path": ens_path})
        if key in self._scratch_ensemble_cache:
            return pickle.loads(self._scratch_ensemble_cache[key])  # nosec

        scratch_ensemble = ScratchEnsemble(ens_name, ens_path).filter("OK")
        self._scratch_ensemble_cache[key] = pickle.dumps(
            scratch_ensemble, pickle.HIGHEST_PROTOCOL
        )

        return scratch_ensemble



from typing import Optional, List, Dict, Any

import pandas as pd
import plotly.graph_objects as go

from webviz_subsurface._abbreviations.number_formatting import si_prefixed
from webviz_subsurface._utils.formatting import printable_int_list
from ._tornado_data import TornadoData


class TornadoBarChart:
    """Creates a plotly bar figure from a TornadoData instance"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        tornado_data: TornadoData,
        plotly_theme: Dict[str, Any],
        figure_height: Optional[int] = None,
        label_options: str = "detailed",
        locked_si_prefix: Optional[int] = None,
        number_format: str = "",
        unit: str = "",
        spaced: bool = True,
        use_true_base: bool = False,
        show_realization_points: bool = True,
    ) -> None:
        self._tornadotable = tornado_data.tornadotable
        self._realtable = self.make_points(tornado_data.real_df)
        self._reference_average = tornado_data.reference_average
        self._plotly_theme = plotly_theme
        self._number_format = number_format
        self._unit = unit
        self._spaced = spaced
        self._locked_si_prefix = locked_si_prefix
        self._locked_si_prefix_relative: Optional[int]
        self._scale = tornado_data.scale
        self._use_true_base = use_true_base
        if self._scale == "Percentage":
            self._unit_x = "%"
            self._locked_si_prefix_relative = 0
        else:
            self._unit_x = self._unit
            self._locked_si_prefix_relative = locked_si_prefix
        self._figure_height = figure_height
        self._label_options = label_options
        self._show_scatter = show_realization_points

    def make_points(self, realdf: pd.DataFrame) -> pd.DataFrame:
        dfs = []
        for sensname in self._tornadotable["sensname"].unique():
            for case in ["high", "low"]:
                df = realdf.loc[
                    realdf["REAL"].isin(
                        self._tornadotable[self._tornadotable["sensname"] == sensname][
                            f"{case}_reals"
                        ].iloc[0]
                    )
                ].copy()
                df["sensname"] = sensname
                df["case"] = case
                dfs.append(df)
        return (
            pd.concat(dfs)
            if dfs
            else pd.DataFrame(columns=["REAL", "sensname", "case"])
        )

    @property
    def figure_height(self) -> Optional[int]:
        """Set height of figure as a function of number of senscases(bars)"""
        return self._figure_height

    def _set_si_prefix(self, value: float) -> str:
        return str(
            si_prefixed(
                value,
                self._number_format,
                self._unit,
                self._spaced,
                self._locked_si_prefix,
            )
        )

    def _set_si_prefix_relative(self, value: float) -> str:
        return str(
            si_prefixed(
                value,
                self._number_format,
                self._unit_x,
                self._spaced,
                self._locked_si_prefix_relative,
            )
        )

    def bar_labels(self, case: str) -> List:

        if self._label_options == "simple":
            return [
                f"<b>{self._set_si_prefix_relative(x)}</b>, "
                for x in self._tornadotable[f"{case}_tooltip"]
            ]
        if self._label_options == "detailed":
            return [
                f"<b>{self._set_si_prefix_relative(x)}</b>, "
                f"True: {self._set_si_prefix(val)}, "
                f"<br><b>Case: {label}</b>, "
                f"Reals: {printable_int_list(reals)}"
                if reals
                else None
                for x, label, val, reals in zip(
                    self._tornadotable[f"{case}_tooltip"],
                    self._tornadotable[f"{case}_label"],
                    self._tornadotable[f"true_{case}"],
                    self._tornadotable[f"{case}_reals"],
                )
            ]
        return []

    @property
    def data(self) -> List:
        return [
            dict(
                type="bar",
                y=self._tornadotable["sensname"],
                x=self._tornadotable["low"],
                name="low",
                base=self._tornadotable["low_base"]
                if not self._use_true_base
                else self._reference_average,
                customdata=self._tornadotable["low_reals"],
                text=self.bar_labels("low"),
                textposition="auto",
                insidetextanchor="middle",
                hoverinfo="none",
                orientation="h",
                marker={"line": {"width": 1.5, "color": "black"}},
            ),
            dict(
                type="bar",
                y=self._tornadotable["sensname"],
                x=self._tornadotable["high"],
                name="high",
                base=self._tornadotable["high_base"]
                if not self._use_true_base
                else self._reference_average,
                customdata=self._tornadotable["high_reals"],
                text=self.bar_labels("high"),
                textposition="auto",
                insidetextanchor="middle",
                hoverinfo="none",
                orientation="h",
                marker={"line": {"width": 1.5, "color": "black"}},
            ),
        ]

    def calculate_scatter_value(self, case_values: pd.Series) -> List:
        if self._use_true_base:
            return case_values
        if self._scale == "Percentage":
            return (
                (case_values - self._reference_average) / self._reference_average
            ) * 100
        return case_values - self._reference_average

    @property
    def scatter_data(self) -> List[Dict]:
        return [
            {
                "type": "scatter",
                "mode": "markers",
                "y": df["sensname"],
                "x": self.calculate_scatter_value(df["VALUE"]),
                "text": df["REAL"],
                "hoverinfo": "none",
                "marker": {
                    "size": 15,
                    "color": self._plotly_theme["layout"]["colorway"][0]
                    if case == "low"
                    else self._plotly_theme["layout"]["colorway"][1],
                },
            }
            for case, df in self._realtable.groupby("case")
        ]

    @property
    def range(self) -> List[float]:
        """Calculate x-axis range so that the reference is centered"""
        max_val = max(self._tornadotable[["low", "high"]].abs().max())
        if self._use_true_base:
            return [
                (self._reference_average - max_val) * 0.95,
                (self._reference_average + max_val) * 1.05,
            ]
        return [-max_val * 1.05, max_val * 1.05]

    @property
    def layout(self) -> Dict:
        _layout: Dict[str, Any] = go.Layout()
        _layout.update(self._plotly_theme["layout"])
        _layout.update(
            {
                "height": self.figure_height,
                "barmode": "overlay",
                "margin": {"l": 0, "r": 0, "b": 20, "t": 0, "pad": 21},
                "xaxis": {
                    "title": self._scale,
                    "range": self.range,
                    "autorange": self._show_scatter or self._tornadotable.empty,
                    "showgrid": False,
                    "zeroline": False,
                    "linecolor": "black",
                    "showline": True,
                    "automargin": True,
                    "side": "top",
                    "tickfont": {"size": 15},
                },
                "yaxis": {
                    "autorange": True,
                    "showgrid": False,
                    "zeroline": False,
                    "showline": False,
                    "automargin": True,
                    "title": None,
                    "dtick": 1,
                    "tickfont": {"size": 15},
                },
                "showlegend": False,
                "hovermode": "y",
                "annotations": [
                    {
                        "x": 0 if not self._use_true_base else self._reference_average,
                        "y": 1.05,
                        "xref": "x",
                        "yref": "paper",
                        "text": f"<b>{self._set_si_prefix(self._reference_average)}</b>"
                        " (Ref avg)",
                        "showarrow": False,
                        "align": "center",
                    }
                ],
                "shapes": [
                    {
                        "type": "line",
                        "line": {"width": 3, "color": "lightgrey"},
                        "x0": 0 if not self._use_true_base else self._reference_average,
                        "x1": 0 if not self._use_true_base else self._reference_average,
                        "y0": 0,
                        "y1": 1,
                        "xref": "x",
                        "yref": "y domain",
                    }
                ],
            }
        )
        return _layout

    @property
    def figure(self) -> go.Figure:
        data = self.data

        fig = go.Figure({"data": data, "layout": self.layout})
        if self._show_scatter:
            fig.update_traces(marker_opacity=0.4, text=None)
            for trace in self.scatter_data:
                fig.add_trace(trace)
        return fig



# pylint: disable=no-name-in-module
from webviz_config.plugins import RftPlotter


def test_rft_plotter(dash_duo, app, shared_settings, testdata_folder) -> None:
    plugin = RftPlotter(
        app,
        shared_settings["HM_SETTINGS"],
        ensembles=shared_settings["HM_ENSEMBLES"],
        formations=testdata_folder
        / "01_drogon_ahm"
        / "realization-0"
        / "iter-0"
        / "share"
        / "results"
        / "tables"
        / "formations_res_only.csv",
        faultlines=testdata_folder
        / "01_drogon_ahm"
        / "realization-0"
        / "iter-0"
        / "share"
        / "results"
        / "polygons"
        / "toptherys--gl_faultlines_extract_postprocess.csv",
    )
    app.layout = plugin.layout
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



from typing import List, Dict, Optional, Iterable, Tuple

from itertools import chain
import pandas as pd
from webviz_config.common_cache import CACHE

from webviz_subsurface._abbreviations.reservoir_simulation import (
    historical_vector,
)
from webviz_subsurface._utils.simulation_timeseries import (
    set_simulation_line_shape_fallback,
    get_simulation_line_shape,
)


class SimulationTimeSeriesModel:
    """Class to process and and visualize ensemble timeseries"""

    def __init__(
        self,
        dataframe: pd.DataFrame,
        metadata: Optional[pd.DataFrame] = None,
        line_shape_fallback: str = "linear",
    ) -> None:

        for column in ["ENSEMBLE", "REAL", "DATE"]:
            if column not in dataframe.columns:
                raise KeyError(f"{column} column is missing from UNSMRY data")

        self._dataframe = dataframe
        self._metadata = metadata
        self.line_shape_fallback = set_simulation_line_shape_fallback(
            line_shape_fallback
        )

        self._vector_names = SimulationTimeSeriesModel._determine_vector_names(
            self._dataframe, self._metadata
        )
        # Strip out vector names where all data is 0
        self._vector_names = (
            SimulationTimeSeriesModel._filter_vector_names_discard_0_columns(
                self._vector_names, self._dataframe
            )
        )

        self._vector_groups = SimulationTimeSeriesModel._split_vectors_by_type(
            self._vector_names
        )

        self._dates = sorted(self._dataframe["DATE"].unique().astype(str))

        # Currently we cannot be sure what data type the DATE column has.
        # Apparently it will have datetime if the source is SMRY files, but
        # may be of type str if the source is CSV files.
        # Try and detect if the data type is string by sampling the first entry
        sample_date = self._dataframe["DATE"].iloc[0]
        self._date_column_is_str = isinstance(sample_date, str)

    @staticmethod
    def _determine_vector_names(
        dataframe: pd.DataFrame, metadata: Optional[pd.DataFrame]
    ) -> List[str]:
        """Determine which vectors we should make available"""
        vecnames = [
            c
            for c in dataframe.columns
            if c not in ["REAL", "ENSEMBLE", "DATE"]
            and not historical_vector(c, metadata, False) in dataframe.columns
        ]
        return vecnames

    @staticmethod
    def _filter_vector_names_discard_0_columns(
        vector_names: Iterable[str], dataframe: pd.DataFrame
    ) -> List[str]:
        """Filter list of vector names by removing vector names where the data is all zeros"""
        non_zero_column_mask = (dataframe != 0).any(axis=0)
        non_zero_column_names = set(dataframe.columns[non_zero_column_mask])
        filtered_vecnames = [n for n in vector_names if n in non_zero_column_names]
        return filtered_vecnames

    @staticmethod
    def _split_vectors_by_type(vector_names: Iterable[str]) -> Dict[str, dict]:
        vtypes = ["Field", "Well", "Region", "Block", "Group", "Connection"]
        vgroups = {}
        for vtype in vtypes:
            vectors = [v for v in vector_names if v.startswith(vtype[0])]
            if vectors:
                shortnames, items = SimulationTimeSeriesModel._vector_subitems(vectors)
                vgroups[vtype] = dict(
                    vectors=vectors, shortnames=shortnames, items=items
                )

        other_vectors = [
            x
            for x in vector_names
            if x
            not in list(
                chain.from_iterable(
                    [vtype_dict["vectors"] for vtype_dict in vgroups.values()]
                )
            )
        ]
        if other_vectors:
            vgroups["Others"] = dict(
                vectors=other_vectors, shortnames=other_vectors, items=[]
            )
        return vgroups

    @staticmethod
    def _vector_subitems(vectors: Iterable[str]) -> Tuple[List[str], List[str]]:
        shortnames = {v.split(":")[0] for v in vectors}
        items = {":".join(v.split(":")[1:]) for v in vectors if len(v.split(":")) > 1}

        return sorted(shortnames), sorted(
            items, key=int if all(item.isdigit() for item in items) else None
        )

    @property
    def dates(self) -> List[str]:
        return self._dates

    @property
    def vectors(self) -> List[str]:
        return self._vector_names

    @property
    def vector_groups(self) -> Dict[str, dict]:
        return self._vector_groups

    @property
    def ensembles(self) -> List[str]:
        return list(self._dataframe["ENSEMBLE"].unique())

    def get_line_shape(self, vector: str) -> str:
        return get_simulation_line_shape(
            line_shape_fallback=self.line_shape_fallback,
            vector=vector,
            smry_meta=self._metadata,
        )

    def get_ensemble_vectors_for_date(
        self, ensemble: str, date: str, vectors: list = None
    ) -> pd.DataFrame:
        vectors = vectors if vectors is not None else self._vector_names

        # If the data type of the DATE column is string, use the passed date argument
        # directly. Otherwise convert to datetime before doing query
        query_date = date if self._date_column_is_str else pd.to_datetime(date)
        df = self._dataframe.loc[
            (self._dataframe["ENSEMBLE"] == ensemble)
            & (self._dataframe["DATE"] == query_date),
            vectors + ["REAL"],
        ]
        return df

    def _add_history_trace(self, dframe: pd.DataFrame, vector: str) -> dict:
        """Renders the history line"""
        df = dframe.loc[
            (dframe["REAL"] == dframe["REAL"].unique()[0])
            & (dframe["ENSEMBLE"] == dframe["ENSEMBLE"].unique()[0])
        ]
        return {
            "line": {"shape": self.get_line_shape(vector)},
            "x": df["DATE"].astype(str),
            "y": df[vector],
            "hovertext": "History",
            "hoverinfo": "y+x+text",
            "name": "History",
            "marker": {"color": "black"},
            "showlegend": True,
        }

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    def add_realization_traces(
        self, ensemble: str, vector: str, real_filter: Optional[list] = None
    ) -> list:
        """Renders line trace for each realization, includes history line if present"""
        dataframe = self._dataframe[self._dataframe["ENSEMBLE"] == ensemble]
        dataframe = (
            dataframe[dataframe["REAL"].isin(real_filter)]
            if real_filter is not None
            else dataframe
        )
        traces = [
            {
                "line": {"shape": self.get_line_shape(vector)},
                "x": list(real_df["DATE"].astype(str)),
                "y": list(real_df[vector]),
                "name": ensemble,
                "customdata": real,
                "legendgroup": ensemble,
                "marker": {"color": "red"},
                "showlegend": real_idx == 0,
            }
            for real_idx, (real, real_df) in enumerate(dataframe.groupby("REAL"))
        ]

        hist_vecname = historical_vector(vector=vector, smry_meta=self._metadata)
        if hist_vecname and hist_vecname in dataframe.columns:
            traces.append(
                self._add_history_trace(
                    dataframe,
                    hist_vecname,
                )
            )

        return traces

    def daterange_for_plot(self, vector: str) -> List[str]:
        date = self._dataframe["DATE"][self._dataframe[vector] != 0]
        return [str(date.min()), str(date.max())]



from typing import List, Dict, Tuple

import numpy as np
import pandas as pd
from dash import html, dcc, Dash, Input, Output, State, ALL
from dash.exceptions import PreventUpdate
import webviz_core_components as wcc
import dash_bootstrap_components as dbc
from webviz_config.utils import calculate_slider_step


class ParameterFilter:
    """Component that can be added to a plugin to filter parameters"""

    def __init__(self, app: Dash, uuid: str, dframe: pd.DataFrame) -> None:
        """
        * **`app`:** The Dash app instance.
        * **`uuid`:** Unique id (use the plugin id).
        * **`dframe`:** Dataframe, of all parameter values in all ensembles"""
        self.app = app
        self._uuid = uuid
        self._dframe = dframe
        self._validate_dframe()
        self._prepare_data()
        self.set_callbacks(app)

    def _validate_dframe(self) -> None:
        for col in ["REAL", "ENSEMBLE"]:
            if col not in self._dframe.columns:
                raise KeyError(f"Required columns {col} not found in dataframe")

    def _prepare_data(self, drop_constants: bool = True) -> None:
        """
        Different data preparations on the parameters, before storing them as an attribute.
        Option to drop parameters with constant values. Prefixes on parameters from GEN_KW
        are removed, in addition parameters with LOG distribution will be kept while the
        other is dropped.
        """

        if drop_constants:
            constant_params = [
                param
                for param in [
                    x for x in self._dframe.columns if x not in ["REAL", "ENSEMBLE"]
                ]
                if len(self._dframe[param].unique()) == 1
            ]
            self._dframe = self._dframe.drop(columns=constant_params)

        # Keep only LOG parameters
        log_params = [
            param.replace("LOG10_", "")
            for param in [
                x for x in self._dframe.columns if x not in ["REAL", "ENSEMBLE"]
            ]
            if param.startswith("LOG10_")
        ]
        self._dframe = self._dframe.drop(columns=log_params)
        self._dframe = self._dframe.rename(
            columns={
                col: f"{col} (log)"
                for col in self._dframe.columns
                if col.startswith("LOG10_")
            }
        )
        # Remove prefix on parameter name added by GEN_KW
        self._dframe = self._dframe.rename(
            columns={
                col: (col.split(":", 1)[1])
                for col in self._dframe.columns
                if (":" in col and col not in ["REAL", "ENSEMBLE"])
            }
        )
        # Drop columns if duplicate names
        self._dframe = self._dframe.loc[:, ~self._dframe.columns.duplicated()]

    @property
    def _ensembles(self) -> List[str]:
        return list(self._dframe["ENSEMBLE"].unique())

    @property
    def is_sensitivity_run(self) -> bool:
        return "SENSCASE" in self._dframe.columns and "SENSNAME" in self._dframe.columns

    @property
    def _constant_parameters(self) -> List[str]:
        return [
            col
            for col in self._dframe.columns[self._dframe.nunique() <= 1]
            if col not in ["REAL", "ENSEMBLE"]
        ]

    @property
    def _range_parameters(self) -> List[str]:
        numeric_df = self._dframe.select_dtypes(include=np.number)
        return [
            col
            for col in numeric_df.columns[numeric_df.nunique() >= 10]
            if col not in ["REAL", "ENSEMBLE"]
        ]

    @property
    def _discrete_parameters(self) -> List[str]:
        return [
            col
            for col in self._dframe.columns
            if col
            not in ["REAL", "ENSEMBLE"]
            + self._constant_parameters
            + self._range_parameters
            + ["SENSNAME", "SENSCASE"]
        ]

    @property
    def _range_sliders(self) -> html.Div:
        return html.Div(
            children=[
                make_range_slider(values=self._dframe[col], name=col, uuid=self._uuid)
                for col in self._range_parameters
            ]
        )

    @property
    def _discrete_selectors(self) -> html.Div:
        return html.Div(
            children=[
                make_discrete_selector(
                    values=self._dframe[col], name=col, uuid=self._uuid
                )
                for col in self._discrete_parameters
            ]
        )

    @property
    def layout(self) -> html.Div:
        return html.Div(
            style={"height": "100%"},
            children=[
                html.Div(
                    style={"height": "90%"},
                    children=[
                        wcc.Header(
                            style={"color": "black"},
                            children="Parameter filter",
                        ),
                        html.Div(
                            style={"overflowY": "auto", "height": "90%"},
                            children=[self._range_sliders, self._discrete_selectors],
                        ),
                    ],
                ),
                self.buttons,
                dcc.Store(
                    id={"id": self._uuid, "type": "data-store"},
                    data=self._initial_store,
                ),
            ],
        )

    @property
    def _initial_store(self) -> Dict[str, List]:
        data = {}
        for ens_name, ens_df in self._dframe.groupby("ENSEMBLE"):
            data[ens_name] = sorted(list(ens_df["REAL"].unique()))
        return data

    @property
    def buttons(self) -> html.Div:
        return html.Div(
            style={"marginTop": "20px"},
            children=[
                dbc.Button(
                    "Reset",
                    style={
                        "width": "48%",
                        "float": "right",
                        "background-color": "white",
                    },
                    className="mr-1",
                    id={"id": self._uuid, "type": "button", "element": "reset"},
                ),
                dbc.Button(
                    "Apply",
                    # style={"padding": "0 20px", "visibility": "hidden"}
                    # if apply_disabled
                    style={
                        "width": "48%",
                        "float": "left",
                        "background-color": "white",
                    },
                    className="mr-1",
                    id={"id": self._uuid, "type": "button", "element": "apply"},
                ),
            ],
        )

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            Output({"id": self._uuid, "type": "data-store"}, "data"),
            Input({"id": self._uuid, "type": "button", "element": "apply"}, "n_clicks"),
            State({"id": self._uuid, "type": "range-slider", "name": ALL}, "value"),
            State({"id": self._uuid, "type": "select", "name": ALL}, "value"),
        )
        def store_selections(n_clicks: int, range_sliders: list, selects: list) -> Dict:
            if not n_clicks:
                raise PreventUpdate

            dframe = self._dframe
            real_dict = {}
            for values, name in zip(range_sliders, self._range_parameters):
                dframe = dframe[
                    (dframe[name] >= values[0]) & (dframe[name] <= values[1])
                ]
            for values, name in zip(selects, self._discrete_parameters):
                dframe = dframe[dframe[name].isin(values)]
            for ens, ens_df in dframe.groupby("ENSEMBLE"):
                real_dict[ens] = list(ens_df["REAL"].unique())

            for ens in self._ensembles:
                if ens not in real_dict.keys():
                    real_dict[ens] = []
            return real_dict

        @app.callback(
            Output({"id": self._uuid, "type": "range-slider", "name": ALL}, "value"),
            Output({"id": self._uuid, "type": "select", "name": ALL}, "value"),
            Output(
                {"id": self._uuid, "type": "button", "element": "apply"}, "n_clicks"
            ),
            Input({"id": self._uuid, "type": "button", "element": "reset"}, "n_clicks"),
            State({"id": self._uuid, "type": "button", "element": "apply"}, "n_clicks"),
        )
        def reset_selections(n_clicks: int, apply_click: int) -> Tuple[List, List, int]:
            if not n_clicks:
                raise PreventUpdate
            range_sliders = [
                [self._dframe[col].min(), self._dframe[col].max()]
                for col in self._range_parameters
            ]
            selects = [
                sorted(list(self._dframe[col].unique()))
                for col in self._discrete_parameters
            ]
            return range_sliders, selects, apply_click


def make_range_slider(values: pd.Series, name: str, uuid: str) -> html.Div:
    return wcc.RangeSlider(
        label=name,
        id={"id": uuid, "type": "range-slider", "name": name},
        min=values.min(),
        max=values.max(),
        step=calculate_slider_step(
            min_value=values.min(),
            max_value=values.max(),
            steps=len(list(values.unique())) - 1,
        ),
        value=[values.min(), values.max()],
        marks={
            str(values.min()): {"label": f"{values.min():.2f}"},
            str(values.max()): {"label": f"{values.max():.2f}"},
        },
        tooltip={"always_visible": False},
    )


def make_discrete_selector(values: pd.Series, name: str, uuid: str) -> html.Div:
    return wcc.SelectWithLabel(
        label=name,
        id={"id": uuid, "type": "select", "name": name},
        options=[{"label": val, "value": val} for val in values.unique()],
        value=sorted(list(values.unique())),
        multi=True,
    )



from dash import html, dcc, dash_table
import plotly.graph_objects as go
import webviz_core_components as wcc
from webviz_subsurface._models import InplaceVolumesModel


def comparison_main_layout(uuid: str) -> html.Div:
    return wcc.Frame(
        color="white",
        highlight=False,
        style={"height": "91vh"},
        children=[
            html.Div(
                style={"margin-bottom": "20px"},
                children=wcc.RadioItems(
                    vertical=False,
                    id={"id": uuid, "element": "display-option"},
                    options=[
                        {
                            "label": "QC plots",
                            "value": "plots",
                        },
                        {
                            "label": "Difference table for selected response",
                            "value": "single-response table",
                        },
                        {
                            "label": "Difference table for multiple responses",
                            "value": "multi-response table",
                        },
                    ],
                    value="plots",
                ),
            ),
            html.Div(id={"id": uuid, "wrapper": "table"}),
        ],
    )


def comparison_qc_plots_layout(
    fig_diff_vs_real: go.Figure,
    fig_corr: go.Figure,
    fig_diff_vs_response: go.Figure,
    barfig: go.Figure,
) -> html.Div:
    return html.Div(
        children=[
            html.Div(
                children=wcc.Graph(
                    config={"displayModeBar": False},
                    style={"height": "22vh"},
                    figure=fig_diff_vs_real,
                )
            ),
            wcc.FlexBox(
                style={"height": "32vh"},
                children=[
                    wcc.FlexColumn(
                        children=wcc.Graph(
                            config={"displayModeBar": False},
                            style={"height": "31vh"},
                            figure=fig_diff_vs_response,
                        )
                    ),
                    wcc.FlexColumn(
                        children=wcc.Graph(
                            config={"displayModeBar": False},
                            style={"height": "31vh"},
                            figure=fig_corr,
                        )
                    ),
                ],
            ),
            wcc.Frame(
                style={"height": "31vh"},
                children=[
                    wcc.Header("Highlighted data"),
                    wcc.Graph(
                        config={"displayModeBar": False},
                        style={"height": "25vh"},
                        figure=barfig,
                    )
                    if barfig is not None
                    else html.Div("No data within highlight criteria"),
                ],
            ),
        ]
    )


def comparison_table_layout(
    table: dash_table, table_type: str, selections: dict, filter_info: str
) -> html.Div:
    if table_type == "single-response table":
        header = f"Table showing differences for {selections['Response']}"
    else:
        diff_mode = "percent" if selections["Diff mode"] == "diff (%)" else "true value"
        header = f"Table showing differences in {diff_mode} for multiple responses"

    return html.Div(
        children=[
            wcc.Header(header),
            html.Div(
                style={"margin-bottom": "30px", "font-weight": "bold"},
                children=[
                    html.Div(f"From {selections['value1']} to {selections['value2']}"),
                    html.Div(
                        f"{filter_info.capitalize()} {selections['filters'][filter_info][0]}"
                    ),
                ],
            ),
            table,
        ]
    )


def settings_layout(uuid: str, tab: str) -> wcc.Selectors:
    return wcc.Selectors(
        label="⚙️ SETTINGS",
        open_details=False,
        children=[
            colorby_selector(uuid, tab),
            axis_focus_selector(uuid, tab),
            remove_zero_responses(uuid, tab),
            remove_non_highlighted_data(uuid, tab),
        ],
    )


def comparison_selections(
    uuid: str, volumemodel: InplaceVolumesModel, tab: str, compare_on: str
) -> html.Div:
    elements = volumemodel.sources if compare_on == "SOURCE" else volumemodel.ensembles
    return html.Div(
        children=[
            wcc.Selectors(
                label="CONTROLS",
                open_details=True,
                children=[
                    source_selector(
                        uuid,
                        tab,
                        label=f"{compare_on.capitalize()} A",
                        selector_label="value1",
                        value=elements[0],
                        elements=elements,
                    ),
                    source_selector(
                        uuid,
                        tab,
                        label=f"{compare_on.capitalize()} B",
                        selector_label="value2",
                        value=elements[1] if compare_on == "SOURCE" else elements[-1],
                        elements=elements,
                    ),
                    response_selector(volumemodel, uuid, tab),
                    group_by_selector(volumemodel, uuid, tab),
                    diff_mode_selector(uuid, tab),
                    highlight_controls(uuid, tab),
                ],
            ),
            settings_layout(uuid, tab),
        ]
    )


def axis_focus_selector(uuid: str, tab: str) -> html.Div:
    return wcc.Checklist(
        id={"id": uuid, "tab": tab, "selector": "Axis focus"},
        options=[{"label": "Focus diff plots on highlighted", "value": "focus"}],
        value=["focus"],
    )


def remove_zero_responses(uuid: str, tab: str) -> html.Div:
    return wcc.Checklist(
        id={"id": uuid, "tab": tab, "selector": "Remove zeros"},
        options=[{"label": "Remove data with no volume", "value": "remove"}],
        value=["remove"],
    )


def remove_non_highlighted_data(uuid: str, tab: str) -> html.Div:
    return wcc.Checklist(
        id={"id": uuid, "tab": tab, "selector": "Remove non-highlighted"},
        options=[
            {"label": "Display only highlighted data in table", "value": "remove"}
        ],
        value=[],
    )


def diff_mode_selector(uuid: str, tab: str) -> html.Div:
    return html.Div(
        style={"margin-top": "10px"},
        children=wcc.RadioItems(
            label="Difference mode",
            id={"id": uuid, "tab": tab, "selector": "Diff mode"},
            options=[
                {"label": "Percent", "value": "diff (%)"},
                {"label": "True value", "value": "diff"},
            ],
            labelStyle={"display": "inline-flex", "margin-right": "5px"},
            value="diff (%)",
        ),
    )


def highlight_controls(uuid: str, tab: str) -> html.Div:
    return html.Div(
        style={"margin-top": "10px"},
        children=[
            html.Label("Data highlight criterias", className="webviz-underlined-label"),
            html.Div(
                children=[
                    wcc.Label("Absolute diff (%) above:"),
                    dcc.Input(
                        id={"id": uuid, "tab": tab, "selector": "Accept value"},
                        type="number",
                        required=True,
                        value=5,
                        persistence=True,
                        persistence_type="session",
                    ),
                ],
            ),
            html.Div(
                children=[
                    wcc.Label("Ignore response values below:"),
                    dcc.Input(
                        id={"id": uuid, "tab": tab, "selector": "Ignore <"},
                        type="number",
                        required=True,
                        value=0,
                        persistence=True,
                        persistence_type="session",
                        debounce=True,
                    ),
                ]
            ),
        ],
    )


def source_selector(
    uuid: str,
    tab: str,
    label: str,
    selector_label: str,
    value: str,
    elements: list,
) -> wcc.Dropdown:

    return wcc.Dropdown(
        label=label,
        id={"id": uuid, "tab": tab, "selector": selector_label},
        options=[{"label": src, "value": src} for src in elements],
        value=value,
        clearable=False,
    )


def response_selector(
    volumemodel: InplaceVolumesModel,
    uuid: str,
    tab: str,
) -> html.Div:
    return html.Div(
        style={"margin-top": "10px"},
        children=wcc.Dropdown(
            id={
                "id": uuid,
                "tab": tab,
                "selector": "Response",
            },
            label="Response",
            options=[{"label": i, "value": i} for i in volumemodel.responses],
            value=volumemodel.volume_columns[0],
            clearable=False,
        ),
    )


def colorby_selector(
    uuid: str,
    tab: str,
) -> html.Div:
    return html.Div(
        style={"margin": "10px 0px"},
        children=wcc.RadioItems(
            label="Color plots on",
            id={"id": uuid, "tab": tab, "selector": "Color by"},
            options=[
                {"label": "Highlighted", "value": "highlighted"},
                {"label": "1st level of investigation", "value": "groups"},
            ],
            labelStyle={"display": "inline-flex", "margin-right": "5px"},
            value="highlighted",
        ),
    )


def group_by_selector(
    volumemodel: InplaceVolumesModel, uuid: str, tab: str
) -> html.Div:
    available_selectors = [
        x
        for x in volumemodel.region_selectors + ["FACIES", "REAL", "FLUID_ZONE"]
        if x in volumemodel.selectors and volumemodel.dataframe[x].nunique() > 1
    ]
    return html.Div(
        style={"margin-top": "10px"},
        children=wcc.Dropdown(
            label="Investigate differences on level",
            id={"id": uuid, "tab": tab, "selector": "Group by"},
            options=[{"label": elm, "value": elm} for elm in available_selectors],
            value=[],
            placeholder="Total",
            multi=True,
            clearable=False,
        ),
    )



from typing import List
from uuid import uuid4
from pathlib import Path

import pandas as pd
import xtgeo
from dash import html, Input, Output, State
from dash.exceptions import PreventUpdate
import webviz_core_components as wcc
from webviz_subsurface_components import LeafletMap
from webviz_config import WebvizPluginABC
from webviz_config import WebvizSettings
from webviz_config.webviz_store import webvizstore
from webviz_config.utils import calculate_slider_step

from webviz_subsurface._models import SurfaceLeafletModel
from .._datainput.grid import load_grid, load_grid_parameter
from .._datainput.surface import get_surface_fence


class SurfaceWithGridCrossSection(WebvizPluginABC):
    """Visualizes surfaces in a map view and grid parameters in a cross section view. \
The cross section is defined by a polyline interactively edited in the map view.

!> This is an experimental plugin exploring how we can visualize 3D grid data in Webviz. \
The performance is currently slow for large grids.

---

* **`gridfile`:** Path to grid geometry (`ROFF` format) (absolute or relative to config file).
* **`gridparameterfiles`:** List of file paths to grid parameters (`ROFF` format) \
 (absolute or relative to config file).
* **`gridparameternames`:** List corresponding to filepaths of displayed parameter names.
* **`surfacefiles`:** List of file paths to surfaces (`irap binary` format) \
 (absolute or relative to config file).
* **`surfacenames`:** List corresponding to file paths of displayed surface names.
* **`zunit`:** z-unit for display.
* **`colors`:** List of hex colors to use. \
Note that apostrophies should be used to avoid that hex colors are read as comments. E.g. \
`'#000000'` for black.

---
**Example files**

* [Gridfile](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
reek_history_match/realization-0/iter-0/share/results/grids/geogrid.roff).
* [One file for gridparameterfiles](https://github.com/equinor/webviz-subsurface-testdata/blob/\
master/reek_history_match/realization-0/iter-0/share/results/grids/geogrid--poro.roff).
* [One file for surfacefiles](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
reek_history_match/realization-0/iter-0/share/results/\
maps/topupperreek--ds_extracted_horizons.gri).

The files above are on a `ROFF binary` format and can be investigated outside `webviz` using \
e.g. [xtgeo](https://xtgeo.readthedocs.io/en/latest/).
"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        app,
        webviz_settings: WebvizSettings,
        gridfile: Path,
        gridparameterfiles: List[Path],
        surfacefiles: List[Path],
        gridparameternames: list = None,
        surfacenames: list = None,
        zunit="depth (m)",
        colors: list = None,
    ):

        super().__init__()
        self.zunit = zunit
        self.gridfile = str(gridfile)
        self.gridparafiles = [str(gridfile) for gridfile in gridparameterfiles]
        self.surfacefiles = [str(surffile) for surffile in surfacefiles]
        if surfacenames is not None:
            if len(surfacenames) != len(surfacefiles):
                raise ValueError(
                    "List of surface names specified should be same length as list of surfacefiles"
                )
            self.surfacenames = surfacenames
        else:
            self.surfacenames = [Path(surfacefile).stem for surfacefile in surfacefiles]
        if gridparameternames is not None:
            if len(gridparameternames) != len(gridparameterfiles):
                raise ValueError(
                    "List of grid parameter names specified should be same length "
                    "as list of gridparameterfiles"
                )
            self.gridparanames = gridparameternames
        else:
            self.gridparanames = [
                Path(gridfile).stem for gridfile in gridparameterfiles
            ]
        self.plotly_theme = webviz_settings.theme.plotly_theme
        self.initial_colors = (
            colors
            if colors is not None
            else [
                "#440154",
                "#482878",
                "#3e4989",
                "#31688e",
                "#26828e",
                "#1f9e89",
                "#35b779",
                "#6ece58",
                "#b5de2b",
                "#fde725",
            ]
        )
        self.uid = uuid4()
        self.set_callbacks(app)

    def ids(self, element):
        """Generate unique id for dom element"""
        return f"{element}-id-{self.uid}"

    @property
    def tour_steps(self):
        return [
            {
                "id": self.ids("layout"),
                "content": (
                    "Plugin to display surfaces and random lines from a 3D grid parameter. "
                ),
            },
            {
                "id": self.ids("surface"),
                "content": ("The visualized surface."),
            },
            {
                "id": self.ids("map-view"),
                "content": (
                    "Map view of the surface. Use the right toolbar to "
                    "draw a random line."
                ),
            },
            {
                "id": self.ids("fence-view"),
                "content": (
                    "Cross section view of the grid parameter along the edited line. "
                    "The view is empty until a random line is drawn in the map view."
                ),
            },
            {
                "id": self.ids("surface-type"),
                "content": (
                    "Display the z-value of the surface (e.g. depth) or "
                    "the grid parameter value where the surface intersect the grid."
                ),
            },
            {
                "id": self.ids("gridparameter"),
                "content": "The visualized grid parameter.",
            },
            {
                "id": self.ids("color-scale"),
                "content": ("Click this button to change colorscale"),
            },
            {
                "id": self.ids("color-values"),
                "content": ("Drag either node of slider to truncate color ranges"),
            },
            {
                "id": self.ids("color-range-btn"),
                "content": (
                    "Click this button to update color slider min/max and reset ranges."
                ),
            },
        ]

    @property
    def layout(self):
        return wcc.FlexBox(
            id=self.ids("layout"),
            children=[
                wcc.Frame(
                    style={"flex": 1},
                    children=[
                        wcc.Selectors(
                            label="Map settings",
                            children=[
                                wcc.Dropdown(
                                    id=self.ids("surface"),
                                    label="Select surface",
                                    options=[
                                        {"label": name, "value": path}
                                        for name, path in zip(
                                            self.surfacenames, self.surfacefiles
                                        )
                                    ],
                                    value=self.surfacefiles[0],
                                    clearable=False,
                                ),
                                wcc.RadioItems(
                                    id=self.ids("surface-type"),
                                    options=[
                                        {
                                            "label": "Display surface z-value",
                                            "value": "surface",
                                        },
                                        {
                                            "label": "Display seismic attribute as z-value",
                                            "value": "attribute",
                                        },
                                    ],
                                    value="surface",
                                ),
                            ],
                        ),
                        wcc.Selectors(
                            label="Intersection settings",
                            children=[
                                wcc.Dropdown(
                                    label="Select grid parameter",
                                    id=self.ids("gridparameter"),
                                    options=[
                                        {"label": name, "value": para}
                                        for name, para in zip(
                                            self.gridparanames, self.gridparafiles
                                        )
                                    ],
                                    value=self.gridparafiles[0],
                                    clearable=False,
                                ),
                                wcc.Label(
                                    children="Set colorscale",
                                ),
                                wcc.ColorScales(
                                    id=self.ids("color-scale"),
                                    colorscale=self.initial_colors,
                                    nSwatches=12,
                                ),
                                wcc.RangeSlider(
                                    label="Set color range",
                                    id=self.ids("color-values"),
                                    tooltip={"always_visible": False},
                                ),
                                html.Button(
                                    id=self.ids("color-range-btn"),
                                    children="Reset Range",
                                ),
                            ],
                        ),
                    ],
                ),
                wcc.Frame(
                    highlight=False,
                    style={
                        "height": "800px",
                        "flex": 3,
                    },
                    children=[
                        LeafletMap(
                            id=self.ids("map-view"),
                            autoScaleMap=True,
                            minZoom=-19,
                            updateMode="update",
                            layers=[],
                            drawTools={
                                "drawMarker": False,
                                "drawPolygon": False,
                                "drawPolyline": True,
                                "position": "topright",
                            },
                            mouseCoords={"position": "bottomright"},
                            colorBar={"position": "bottomleft"},
                            switch={
                                "value": False,
                                "disabled": False,
                                "label": "Hillshading",
                            },
                        ),
                    ],
                ),
                html.Div(
                    style={"flex": 3, "height": "800px"},
                    children=[
                        wcc.Graph(
                            config={"displayModeBar": False}, id=self.ids("fence-view")
                        ),
                    ],
                ),
            ],
        )

    def set_callbacks(self, app):
        @app.callback(
            Output(self.ids("map-view"), "layers"),
            [
                Input(self.ids("surface"), "value"),
                Input(self.ids("surface-type"), "value"),
                Input(self.ids("gridparameter"), "value"),
                Input(self.ids("color-values"), "value"),
                Input(self.ids("map-view"), "switch"),
            ],
        )
        def _render_surface(
            surfacepath,
            surface_type,
            gridparameter,
            color_values,
            hillshade,
        ):

            surface = xtgeo.surface_from_file(get_path(surfacepath))
            min_val = None
            max_val = None

            if surface_type == "attribute":
                min_val = color_values[0] if color_values else None
                max_val = color_values[1] if color_values else None
                grid = load_grid(get_path(self.gridfile))
                gridparameter = load_grid_parameter(grid, get_path(gridparameter))
                surface.slice_grid3d(grid, gridparameter)

            return [
                SurfaceLeafletModel(
                    surface,
                    name="surface",
                    clip_min=min_val,
                    clip_max=max_val,
                    apply_shading=hillshade.get("value", False),
                ).layer
            ]

        @app.callback(
            Output(self.ids("fence-view"), "figure"),
            [
                Input(self.ids("map-view"), "polyline_points"),
                Input(self.ids("gridparameter"), "value"),
                Input(self.ids("surface"), "value"),
                Input(self.ids("color-values"), "value"),
                Input(self.ids("color-scale"), "colorscale"),
            ],
        )
        def _render_fence(coords, gridparameter, surfacepath, color_values, colorscale):
            if not coords:
                raise PreventUpdate
            grid = load_grid(get_path(self.gridfile))
            gridparameter = load_grid_parameter(grid, get_path(gridparameter))
            fence = get_fencespec(coords)
            hmin, hmax, vmin, vmax, values = grid.get_randomline(
                fence, gridparameter, zincrement=0.5
            )

            surface = xtgeo.surface_from_file(get_path(surfacepath))
            s_arr = get_surface_fence(fence, surface)
            return make_heatmap(
                values,
                s_arr=s_arr,
                theme=self.plotly_theme,
                s_name=self.surfacenames[self.surfacefiles.index(surfacepath)],
                colorscale=colorscale,
                xmin=hmin,
                xmax=hmax,
                ymin=vmin,
                ymax=vmax,
                zmin=color_values[0],
                zmax=color_values[1],
                xaxis_title="Distance along polyline",
                yaxis_title=self.zunit,
            )

        @app.callback(
            [
                Output(self.ids("color-values"), "min"),
                Output(self.ids("color-values"), "max"),
                Output(self.ids("color-values"), "value"),
                Output(self.ids("color-values"), "step"),
            ],
            [Input(self.ids("color-range-btn"), "n_clicks")],
            [State(self.ids("gridparameter"), "value")],
        )
        def _update_color_slider(_clicks, gridparameter):
            grid = load_grid(get_path(self.gridfile))
            gridparameter = load_grid_parameter(grid, get_path(gridparameter))

            minv = float(f"{gridparameter.values.min():2f}")
            maxv = float(f"{gridparameter.values.max():2f}")
            value = [minv, maxv]
            step = calculate_slider_step(minv, maxv, steps=100)
            return minv, maxv, value, step

    def add_webvizstore(self):
        return [
            (get_path, [{"path": fn}])
            for fn in [self.gridfile] + self.gridparafiles + self.surfacefiles
        ]


# pylint: disable=too-many-arguments, too-many-locals
def make_heatmap(
    arr,
    s_arr,
    theme,
    s_color="black",
    s_name=None,
    height=800,
    zmin=None,
    zmax=None,
    xmin=None,
    xmax=None,
    ymin=None,
    ymax=None,
    colorscale=None,
    uirevision=None,
    showscale=True,
    reverse_y=True,
    text=None,
    title=None,
    yaxis_title=None,
    xaxis_title=None,
):

    x_inc = (xmax - xmin) / arr.shape[1]
    y_inc = (ymax - ymin) / arr.shape[0]
    colors = (
        [[i / (len(colorscale) - 1), color] for i, color in enumerate(colorscale)]
        if colorscale
        else "RdBu"
    )
    layout = {}
    layout.update(theme["layout"])
    layout.update(
        {
            "height": height,
            "title": title,
            "uirevision": uirevision,
            "margin": {"b": 50, "t": 50, "r": 0},
            "yaxis": {
                "title": yaxis_title,
                "autorange": "reversed" if reverse_y else None,
            },
            "xaxis": {"title": xaxis_title},
        }
    )
    return {
        "data": [
            {
                "type": "heatmap",
                "name": "seismic",
                "text": text,
                "z": arr.tolist(),
                "x0": xmin,
                "xmax": xmax,
                "dx": x_inc,
                "y0": ymin,
                "ymax": ymax,
                "dy": y_inc,
                "zsmooth": "best",
                "showscale": showscale,
                "colorscale": colors,
                "zmin": zmin,
                "zmax": zmax,
            },
            {
                "type": "line",
                "y": s_arr[:, 2],
                "x": s_arr[:, 3],
                "name": s_name,
                "marker": {"color": s_color},
            },
        ],
        "layout": layout,
    }


@webvizstore
def get_path(path) -> Path:
    return Path(path)


def get_fencespec(coords):
    """Create a XTGeo fence spec from polyline coordinates"""
    poly = xtgeo.Polygons()
    poly.dataframe = pd.DataFrame(
        [
            {
                "X_UTME": c[1],
                "Y_UTMN": c[0],
                "Z_TVDSS": 0,
                "POLY_ID": 1,
                "NAME": "polyline",
            }
            for c in coords
        ]
    )
    return poly.get_fence(asnumpy=True)



#!/usr/bin/env python
"""Exctracts connection status history for each compdat connection that
is included in the summary data on the form CPI:WELL,I,J,K. One line is
added to the export every time a connection changes status. It is OPEN when
CPI>0 and SHUT when CPI=0. The earliest date for any connection will be OPEN,
i.e a cell can not be SHUT before it has been OPEN. This means that any cells
that are always SHUT will be excluded.

The output data set is very sparse compared to the CPI summary data.
"""
import argparse
import re
from typing import List, Tuple, Any, Set
from pathlib import Path

import numpy as np
import pandas as pd
from ecl.summary import EclSum, EclSumKeyWordVector
from webviz_subsurface.plugins._well_completions import WELL_CONNECTION_STATUS_FILE


DESCRIPTION: str = """
Export connection status data on sparse form from CPI summary data.
"""
CATEGORY: str = "utility.eclipse"
EXAMPLES: str = """
Extracts well connection status history from the .UNSMRY file by running
this in the ert workflow:

    FORWARD_MODEL WELL_CONNECTION_STATUS(<ECLBASE>=eclipse/model/SOME)

The default output file is share/results/tables/well_connection_status.parquet
"""


def _get_parser() -> argparse.ArgumentParser:
    """Setup parser for command line options"""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=DESCRIPTION,
    )
    parser.add_argument(
        "eclbase",
        type=Path,
        help="Eclipse base name",
    )
    parser.add_argument(
        "--output",
        type=Path,
        help="Output file",
        default=WELL_CONNECTION_STATUS_FILE,
    )
    return parser


def _get_status_changes(
    dates: np.ndarray, conn_values: np.ndarray
) -> List[Tuple[Any, str]]:
    """Extracts the status history of a single connection as a list of tuples
    on the form (date, status)
    """
    status_changes = []
    prev_value = 0
    for date, value in zip(dates, conn_values):
        if value > 0 and prev_value == 0:
            status_changes.append((date, "OPEN"))
        elif prev_value > 0 and value == 0:
            status_changes.append((date, "SHUT"))
        prev_value = value
    return status_changes


def _extract_well_connection_status(filename: Path) -> pd.DataFrame:
    # pylint: disable=too-many-locals
    """Exctracts well connection status history for each compdat connection that
    is included in the summary data on the form CPI:WELL,I,J,K.

    From the CPI time series it is possible to extract the status of the connection
    because it is 0 when the connection is SHUT and >0 when the connection is open.

    The output from this function is one row for every time a connection changes
    status. The earliest date for any connection will be OPEN, i.e a cell can not
    be SHUT before it has been OPEN. This means that any cells that are always SHUT
    will not be included in the export.
    """

    eclsum = EclSum(str(filename), include_restart=False, lazy_load=False)
    column_names: Set[str] = set(EclSumKeyWordVector(eclsum, add_keywords=True))
    np_dates_ms = eclsum.numpy_dates

    cpi_columns = [
        col
        for col in column_names
        if re.match("^CPI:[A-Z0-9_-]{1,8}:[0-9]+,[0-9]+,[0-9]+$", col)
    ]
    df = pd.DataFrame(columns=["DATE", "WELL", "I", "J", "K", "OP/SH"])

    for col in cpi_columns:
        colsplit = col.split(":")
        well = colsplit[1]
        i, j, k = colsplit[2].split(",")

        vector = eclsum.numpy_vector(col)

        status_changes = _get_status_changes(np_dates_ms, vector)
        for date, status in status_changes:
            df.loc[df.shape[0]] = [date, well, i, j, k, status]

    return df


def main() -> None:
    """Entry point from command line"""
    parser = _get_parser()
    args = parser.parse_args()

    df = _extract_well_connection_status(args.eclbase.with_suffix(".UNSMRY"))

    # Create the output folder if it doesn't exist
    args.output.parent.mkdir(parents=True, exist_ok=True)
    # Write to file
    df.to_parquet(args.output, index=False)


if __name__ == "__main__":
    main()



# pylint: disable=too-many-lines
from typing import List, Dict, Union, Tuple, Callable, Optional
import sys
from pathlib import Path
import json
import datetime
import copy
import warnings

import numpy as np
import pandas as pd
from plotly.subplots import make_subplots
import dash
from dash import html, dcc, Dash, Input, Output, State
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
import webviz_core_components as wcc
from webviz_config import WebvizPluginABC, EncodedFile
from webviz_config import WebvizSettings
from webviz_config.webviz_assets import WEBVIZ_ASSETS
from webviz_config.webviz_store import webvizstore
from webviz_config.common_cache import CACHE
import webviz_subsurface_components as wsc
from webviz_subsurface_components import (
    ExpressionInfo,
    ExternalParseData,
)
import webviz_subsurface

from webviz_subsurface._models import EnsembleSetModel
from webviz_subsurface._models import caching_ensemble_set_model_factory
from .._abbreviations.reservoir_simulation import (
    simulation_vector_description,
    simulation_unit_reformat,
    historical_vector,
)
from .._utils.simulation_timeseries import (
    set_simulation_line_shape_fallback,
    get_simulation_line_shape,
    calc_series_statistics,
    add_statistics_traces,
    render_hovertemplate,
    date_to_interval_conversion,
    check_and_format_observations,
)
from .._utils.fanchart_plotting import (
    FanchartData,
    get_fanchart_traces,
    FreeLineData,
    MinMaxData,
    LowHighData,
)
from .._utils.unique_theming import unique_colors
from .._datainput.from_timeseries_cumulatives import (
    calc_from_cumulatives,
    rename_vec_from_cum,
)

from .._utils.vector_selector import (
    add_vector_to_vector_selector_data,
    is_vector_name_in_vector_selector_data,
)
from .._utils.vector_calculator import (
    expressions_from_config,
    get_calculated_units,
    get_calculated_vector_df,
    get_expression_from_name,
    get_selected_expressions,
    validate_predefined_expression,
    get_custom_vector_definitions_from_expressions,
)


def _check_plugin_options(options: Optional[dict]) -> Optional[Tuple[str, str]]:
    if options:
        if "vector1" in options or "vector2" in options or "vector3" in options:
            return (
                "Please use 'vectors' instead of 'vector1/2/3' in this plugin's options.",
                "Single vector options ('vector1', 'vector2', 'vector3')"
                " have been replaced with a vectors list.",
            )
    return None


# pylint: disable = too-many-instance-attributes
class ReservoirSimulationTimeSeries(WebvizPluginABC):
    """Visualizes reservoir simulation time series data for FMU ensembles.

**Features**
* Visualization of realization time series as line charts.
* Visualization of ensemble time series statistics as line or fan charts.
* Visualization of single date ensemble statistics as histograms.
* Calculation and visualization of delta ensembles.
* Calculation and visualization of average rates and cumulatives over a specified time interval.
* Download of visualized data to csv files (except histogram data).

---
**Two main options for input data: Aggregated and read from UNSMRY.**

**Using aggregated data**
* **`csvfile`:** Aggregated csv file with `REAL`, `ENSEMBLE`, \
    `DATE` and vector columns.

**Using simulation time series data directly from `UNSMRY` files**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize.
* **`column_keys`:** List of vectors to extract. If not given, all vectors \
    from the simulations will be extracted. Wild card asterisk `*` can be used.
* **`sampling`:** Time separation between extracted values. Can be e.g. `monthly` (default) or \
    `yearly`.

**Common optional settings for both input options**
* **`obsfile`**: File with observations to plot together with the relevant time series. \
(absolute path or relative to config file).
* **`options`:** Options to initialize plots with:
    * `vector1` : First vector to display
    * `vector2` : Second vector to display
    * `vector3` : Third vector to display
    * `visualization` : `realizations`, `statistics` or `fanchart`
    * `date` : Date to show by default in histograms
* **`line_shape_fallback`:** Fallback interpolation method between points. Vectors identified as \
    rates or phase ratios are always backfilled, vectors identified as cumulative (totals) are \
    always linearly interpolated. The rest use the fallback.
    Supported options:
    * `linear` (default)
    * `backfilled`
    * `hv`, `vh`, `hvh`, `vhv` and `spline` (regular Plotly options).

**Calculated vector expressions**
* **`predefined_expressions`:** yaml file with pre-defined expressions

---

?> Vectors that are identified as historical vectors (e.g. FOPTH is the history of FOPT) will \
be plotted together with their non-historical counterparts as reference lines, and they are \
therefore not selectable as vectors to plot initially.

?> The `obsfile` is a common (optional) file for all ensembles, which can be \
converted from e.g. ERT and ResInsight formats using the [fmuobs]\
(https://equinor.github.io/subscript/scripts/fmuobs.html) script. \
[An example of the format can be found here]\
(https://github.com/equinor/webviz-subsurface-testdata/blob/master/reek_history_match/share/\
observations/observations.yml).

!> It is **strongly recommended** to keep the data frequency to a regular frequency (like \
`monthly` or `yearly`). This applies to both csv input and when reading from `UNSMRY` \
(controlled by the `sampling` key). This is because the statistics and fancharts are calculated \
per DATE over all realizations in an ensemble, and the available dates should therefore not \
differ between individual realizations of an ensemble.

**Using aggregated data**

The `csvfile` must have columns `ENSEMBLE`, `REAL` and `DATE` in addition to the individual
vectors.
* [Example of aggregated file]\
(https://github.com/equinor/webviz-subsurface-testdata/blob/master/aggregated_data/smry.csv).

**Using simulation time series data directly from `.UNSMRY` files**

Vectors are extracted automatically from the `UNSMRY` files in the individual realizations,
using the `fmu-ensemble` library.

?> Using the `UNSMRY` method will also extract metadata like units, and whether the vector is a \
rate, a cumulative, or historical. Units are e.g. added to the plot titles, while rates and \
cumulatives are used to decide the line shapes in the plot. Aggregated data may on the other \
speed up the build of the app, as processing of `UNSMRY` files can be slow for large models. \
Using this method is required to use the average rate and interval cumulative functionalities, \
as they require identification of vectors that are cumulatives.

!> The `UNSMRY` files are auto-detected by `fmu-ensemble` in the `eclipse/model` folder of the \
individual realizations. You should therefore not have more than one `UNSMRY` file in this \
folder, to avoid risk of not extracting the right data.
"""

    ENSEMBLE_COLUMNS = ["REAL", "ENSEMBLE", "DATE"]

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        csvfile: Path = None,
        ensembles: list = None,
        obsfile: Path = None,
        column_keys: list = None,
        sampling: str = "monthly",
        options: dict = None,
        predefined_expressions: str = None,
        line_shape_fallback: str = "linear",
    ):

        super().__init__()

        WEBVIZ_ASSETS.add(
            Path(webviz_subsurface.__file__).parent
            / "_assets"
            / "css"
            / "block_options.css"
        )

        # Temporary css, pending on new wcc modal component.
        # See: https://github.com/equinor/webviz-core-components/issues/163
        WEBVIZ_ASSETS.add(
            Path(webviz_subsurface.__file__).parent / "_assets" / "css" / "modal.css"
        )

        self.csvfile = csvfile
        self.obsfile = obsfile
        self.time_index = sampling
        self.column_keys = column_keys
        if csvfile and ensembles:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile" or "ensembles"'
            )
        self.observations = {}
        if obsfile:
            self.observations = check_and_format_observations(get_path(self.obsfile))

        self.smry: pd.DataFrame
        self.smry_meta: Union[pd.DataFrame, None]

        if csvfile:
            self.smry = read_csv(csvfile)
            self.smry_meta = None
            # Check of time_index for data to use in resampling. Quite naive as it only checks for
            # unique values of the DATE column, and not per realization.
            #
            # Currently not necessary as we don't allow resampling for average rates and intervals
            # unless we have metadata, which csvfile input currently doesn't support.
            # See: https://github.com/equinor/webviz-subsurface/issues/402
            self.time_index = pd.infer_freq(
                sorted(pd.to_datetime(self.smry["DATE"]).unique())
            )
        elif ensembles:
            self.emodel: EnsembleSetModel = (
                caching_ensemble_set_model_factory.get_or_create_model(
                    ensemble_paths={
                        ens: webviz_settings.shared_settings["scratch_ensembles"][ens]
                        for ens in ensembles
                    },
                    time_index=self.time_index,
                    column_keys=self.column_keys,
                )
            )
            self.smry = self.emodel.get_or_load_smry_cached()
            self.smry_meta = self.emodel.load_smry_meta()
        else:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile" or "ensembles"'
            )
        if any(col.startswith(("AVG_", "INTVL_")) for col in self.smry.columns):
            raise ValueError(
                "Your data set includes time series vectors which have names starting with"
                "'AVG_' and/or 'INTVL_'. These prefixes are not allowed, as they are used"
                "internally in the plugin."
            )
        self.smry_cols: List[str] = [
            c
            for c in self.smry.columns
            if c not in ReservoirSimulationTimeSeries.ENSEMBLE_COLUMNS
            and historical_vector(c, self.smry_meta, False) not in self.smry.columns
        ]

        self.vector_data: list = []
        for vec in self.smry_cols:
            split = vec.split(":")
            add_vector_to_vector_selector_data(
                self.vector_data, vec, simulation_vector_description(split[0])
            )

            if (
                self.smry_meta is not None
                and self.smry_meta.is_total[vec]
                and self.time_index is not None
            ):
                # Get the likely name for equivalent rate vector and make dropdown options.
                # Requires that the time_index was either defined or possible to infer.
                avgrate_vec = rename_vec_from_cum(vector=vec, as_rate=True)
                interval_vec = rename_vec_from_cum(vector=vec, as_rate=False)

                avgrate_split = avgrate_vec.split(":")
                interval_split = interval_vec.split(":")

                add_vector_to_vector_selector_data(
                    self.vector_data,
                    avgrate_vec,
                    f"{simulation_vector_description(avgrate_split[0])} ({avgrate_vec})",
                )
                add_vector_to_vector_selector_data(
                    self.vector_data,
                    interval_vec,
                    f"{simulation_vector_description(interval_split[0])} ({interval_vec})",
                )

        self.ensembles = list(self.smry["ENSEMBLE"].unique())
        self.theme = webviz_settings.theme

        self.plot_options = options if options else {}
        if "vectors" not in self.plot_options:
            self.plot_options["vectors"] = []
        for vector in [
            vector
            for vector in ["vector1", "vector2", "vector3"]
            if vector in self.plot_options
        ]:
            self.plot_options["vectors"].append(self.plot_options[vector])
        self.plot_options["vectors"] = self.plot_options["vectors"][:3]

        self.plot_options["date"] = (
            str(self.plot_options.get("date"))
            if self.plot_options.get("date")
            else str(max(self.smry["DATE"]))
        )
        self.line_shape_fallback = set_simulation_line_shape_fallback(
            line_shape_fallback
        )

        # Retreive predefined expressions from configuration and validate
        self.predefined_expressions_path = (
            None
            if predefined_expressions is None
            else webviz_settings.shared_settings["predefined_expressions"][
                predefined_expressions
            ]
        )
        self.predefined_expressions = expressions_from_config(
            get_path(self.predefined_expressions_path)
            if self.predefined_expressions_path
            else None
        )
        for expression in self.predefined_expressions:
            valid, message = validate_predefined_expression(
                expression, self.vector_data
            )
            if not valid:
                warnings.warn(message)
            expression["isValid"] = valid

        # Create initial vector selector data
        self.initial_vector_selector_data = copy.deepcopy(self.vector_data)
        self._add_expressions_to_vector_data(
            self.initial_vector_selector_data, self.predefined_expressions
        )

        # Check if initially plotted vectors exist in data, raise ValueError if not.
        missing_vectors = (
            [
                value
                for value in self.plot_options["vectors"]
                if value not in self.smry_cols
            ]
            if "vectors" in self.plot_options
            else []
        )
        if missing_vectors:
            raise ValueError(
                f"Cannot find: {', '.join(missing_vectors)} to plot initially in "
                "ReservoirSimulationTimeSeries. Check that the vectors exist in your data, and "
                "that they are not missing in a non-default column_keys list in the yaml config "
                "file."
            )
        self.allow_delta = len(self.ensembles) > 1
        self.set_callbacks(app)

    @staticmethod
    def _add_expression(
        expression_data: list,
        name: str,
        description: Optional[str] = None,
    ) -> None:
        description_str = description if description is not None else ""
        add_vector_to_vector_selector_data(
            vector_selector_data=expression_data,
            vector=name,
            description=description_str,
            description_at_last_node=True,
        )

    @property
    def ens_colors(self) -> dict:
        return unique_colors(self.ensembles, self.theme)

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.uuid("layout"),
                "content": "Dashboard displaying reservoir simulation time series.",
            },
            {
                "id": self.uuid("graph"),
                "content": (
                    "Visualization of selected time series. "
                    "Different options can be set in the menu to the left."
                ),
            },
            {
                "id": self.uuid("ensemble"),
                "content": (
                    "Display time series from one or several ensembles. "
                    "Different ensembles will be overlain in the same plot."
                ),
            },
            {
                "id": self.uuid("vectors"),
                "content": (
                    "Display up to three different time series. "
                    "Each time series will be visualized in a separate plot. "
                    "Vectors prefixed with AVG_ and INTVL_ are calculated in the fly "
                    "from cumulative vectors, providing average rates and interval cumulatives "
                    "over a time interval that can be defined in the menu."
                ),
            },
            {
                "id": self.uuid("visualization"),
                "content": (
                    "Choose between different visualizations. 1. Show time series as "
                    "individual lines per realization. 2. Show statistical lines per "
                    "ensemble. 3. Show statistical fanchart per ensemble"
                    "per date. Select a date by clicking in the plot."
                ),
            },
            {
                "id": self.uuid("options"),
                "content": (
                    "Various plot options: Whether to include history trace, add histogram, "
                    "and which statistics to show if statistical lines is chosen as visualization."
                ),
            },
            {
                "id": self.uuid("cum_interval"),
                "content": (
                    "Defines the time interval the average rates (prefixed AVG_) and interval "
                    "cumulatives (prefixed INTVL_) are calculated over. Disabled unless at least "
                    "one time series dependent on the interval setting is chosen."
                    "The option might be completely hidden if the data input does not support "
                    "calculation from cumulatives."
                ),
            },
            {
                "id": self.uuid("vector_calculator_selector"),
                "content": (
                    "Create mathematical expressions with provided vector time series. "
                    "Parsing of the mathematical expression is handled and will give feedback "
                    "when entering invalid expressions. "
                    "The expressions are calculated on the fly and can be selected among the time "
                    "series to be shown in the plots."
                ),
            },
        ]

    @staticmethod
    def set_grid_layout(columns: str, padding: int = 0) -> Dict[str, str]:
        return {
            "display": "grid",
            "alignContent": "space-around",
            "justifyContent": "space-between",
            "gridTemplateColumns": f"{columns}",
            "padding": f"{padding}px",
        }

    @property
    def time_interval_options(self) -> List[str]:
        if self.time_index == "daily":
            return ["daily", "monthly", "yearly"]
        if self.time_index == "monthly":
            return ["monthly", "yearly"]
        if self.time_index == "yearly":
            return ["yearly"]
        return []

    @property
    def delta_layout(self) -> html.Div:
        show_delta = "block" if self.allow_delta else "none"
        return html.Div(
            children=[
                html.Div(
                    style={"display": show_delta},
                    children=wcc.RadioItems(
                        label="Mode",
                        id=self.uuid("mode"),
                        style={"marginBottom": "0.5vh"},
                        options=[
                            {
                                "label": "Individual ensembles",
                                "value": "ensembles",
                            },
                            {
                                "label": "Delta between ensembles",
                                "value": "delta_ensembles",
                            },
                        ],
                        value="ensembles",
                    ),
                ),
                wcc.Dropdown(
                    wrapper_id=self.uuid("show_ensembles"),
                    label="Selected ensembles",
                    id=self.uuid("ensemble"),
                    clearable=False,
                    multi=True,
                    options=[{"label": i, "value": i} for i in self.ensembles],
                    value=[self.ensembles[0]],
                ),
                html.Div(
                    id=self.uuid("calc_delta"),
                    style={"display": "none"},
                    children=[
                        wcc.Label("Selected ensemble delta (A-B):"),
                        wcc.FlexBox(
                            children=[
                                wcc.FlexColumn(
                                    min_width="100px",
                                    children=wcc.Dropdown(
                                        label="Ensemble A",
                                        id=self.uuid("base_ens"),
                                        clearable=False,
                                        options=[
                                            {"label": i, "value": i}
                                            for i in self.ensembles
                                        ],
                                        value=self.ensembles[0],
                                    ),
                                ),
                                wcc.FlexColumn(
                                    min_width="100px",
                                    children=wcc.Dropdown(
                                        label="Ensemble B",
                                        id=self.uuid("delta_ens"),
                                        clearable=False,
                                        options=[
                                            {"label": i, "value": i}
                                            for i in self.ensembles
                                        ],
                                        value=self.ensembles[-1],
                                    ),
                                ),
                            ],
                        ),
                    ],
                ),
            ],
        )

    @property
    def from_cumulatives_layout(self) -> html.Div:
        return html.Div(
            style=(
                {}
                if len(self.time_interval_options) > 0 and self.smry_meta is not None
                else {"display": "none"}
            ),
            children=[
                wcc.Label("Calculated from cumulatives:"),
                wcc.Label(
                    "Average (AVG_) and interval (INTVL_) time series",
                    style={"font-style": "italic"},
                ),
                html.Div(
                    wcc.RadioItems(
                        id=self.uuid("cum_interval"),
                        className="block-options",
                        options=[
                            {
                                "label": (f"{i.lower().capitalize()}"),
                                "value": i,
                                "disabled": False,
                            }
                            for i in self.time_interval_options
                        ],
                        value=self.time_index,
                    ),
                ),
            ],
        )

    @property
    def modal_vector_calculator_layout(self) -> html.Div:
        return dbc.Modal(
            style={"marginTop": "20vh", "width": "1300px"},
            children=[
                dbc.ModalHeader("Vector Calculator"),
                dbc.ModalBody(
                    html.Div(
                        id=self.uuid("vector_calculator_modal_body"),
                        children=[
                            wsc.VectorCalculator(
                                id=self.uuid("vector_calculator"),
                                vectors=self.vector_data,
                                expressions=self.predefined_expressions,
                            )
                        ],
                    ),
                ),
            ],
            id=self.uuid("modal_vector_calculator"),
            size="lg",
            centered=True,
        )

    @property
    def layout(self) -> wcc.FlexBox:
        return wcc.FlexBox(
            id=self.uuid("layout"),
            children=[
                wcc.FlexColumn(
                    children=wcc.Frame(
                        style={"height": "90vh"},
                        children=[
                            wcc.Selectors(
                                label="Ensembles", children=[self.delta_layout]
                            ),
                            wcc.Selectors(
                                label="Time series",
                                children=wsc.VectorSelector(
                                    id=self.uuid("vectors"),
                                    maxNumSelectedNodes=3,
                                    data=self.initial_vector_selector_data,
                                    placeholder="Add new vector...",
                                    persistence=True,
                                    persistence_type="session",
                                    selectedTags=self.plot_options.get(
                                        "vectors", [self.smry_cols[0]]
                                    ),
                                    numSecondsUntilSuggestionsAreShown=0.5,
                                    lineBreakAfterTag=True,
                                    customVectorDefinitions=(
                                        get_custom_vector_definitions_from_expressions(
                                            self.predefined_expressions
                                        )
                                    ),
                                ),
                            ),
                            wcc.Selectors(
                                label="Visualization",
                                id=self.uuid("visualization"),
                                children=[
                                    wcc.RadioItems(
                                        id=self.uuid("statistics"),
                                        options=[
                                            {
                                                "label": "Individual realizations",
                                                "value": "realizations",
                                            },
                                            {
                                                "label": "Statistical lines",
                                                "value": "statistics",
                                            },
                                            {
                                                "label": "Statistical fanchart",
                                                "value": "fanchart",
                                            },
                                        ],
                                        value=self.plot_options.get(
                                            "visualization", "statistics"
                                        ),
                                    ),
                                ],
                            ),
                            wcc.Selectors(
                                label="Options",
                                id=self.uuid("options"),
                                children=[
                                    wcc.Checklist(
                                        id=self.uuid("trace_options"),
                                        options=[
                                            {"label": val, "value": val}
                                            for val in ["History", "Histogram"]
                                        ],
                                        value=["History"],
                                    ),
                                    html.Div(
                                        id=self.uuid("view_stat_options"),
                                        style={"display": "block"}
                                        if "statistics"
                                        in self.plot_options.get("visualization", "")
                                        else {"display": "none"},
                                        children=[
                                            wcc.Checklist(
                                                id=self.uuid("stat_options"),
                                                options=[
                                                    {"label": val, "value": val}
                                                    for val in [
                                                        "Mean",
                                                        "P10 (high)",
                                                        "P50 (median)",
                                                        "P90 (low)",
                                                        "Maximum",
                                                        "Minimum",
                                                    ]
                                                ],
                                                value=[
                                                    "Mean",
                                                    "P10 (high)",
                                                    "P90 (low)",
                                                ],
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                            wcc.Selectors(
                                label="Calculations",
                                children=[self.from_cumulatives_layout],
                            ),
                            wcc.Selectors(
                                label="Vector Calculator",
                                id=self.uuid("vector_calculator_selector"),
                                open=False,
                                children=[
                                    dbc.Button(
                                        "Vector Calculator",
                                        id=self.uuid("vector_calculator_open_btn"),
                                    ),
                                ],
                            ),
                        ],
                    )
                ),
                self.modal_vector_calculator_layout,
                wcc.FlexColumn(
                    flex=4,
                    children=[
                        wcc.Frame(
                            style={"height": "90vh"},
                            highlight=False,
                            color="white",
                            children=wcc.Graph(
                                style={"height": "85vh"},
                                id=self.uuid("graph"),
                            ),
                        ),
                        dcc.Store(
                            id=self.uuid("date"),
                            storage_type="session",
                            data=json.dumps(self.plot_options.get("date", None)),
                        ),
                    ],
                ),
                dcc.Store(
                    id=self.uuid("vector_calculator_expressions_modal_open"),
                    data=self.predefined_expressions,
                ),
                dcc.Store(
                    id=self.uuid("vector_calculator_expressions"),
                    data=self.predefined_expressions,
                ),
            ],
        )

    @staticmethod
    def _get_valid_vector_selections(
        vector_data: list,
        selected_vectors: List[str],
        new_expressions: List[ExpressionInfo],
        existing_expressions: List[ExpressionInfo],
    ) -> List[str]:
        valid_selections: List[str] = []
        for vector in selected_vectors:
            new_vector: Optional[str] = vector

            # Get id if vector is among existing expressions
            dropdown_id = next(
                (elm["id"] for elm in existing_expressions if elm["name"] == vector),
                None,
            )
            # Find id among new expressions to get new/edited name
            if dropdown_id:
                new_vector = next(
                    (
                        elm["name"]
                        for elm in new_expressions
                        if elm["id"] == dropdown_id
                    ),
                    None,
                )

            # Append if vector name exist among data
            if new_vector is not None and is_vector_name_in_vector_selector_data(
                new_vector, vector_data
            ):
                valid_selections.append(new_vector)

        return valid_selections

    def _add_expressions_to_vector_data(
        self, vector_data: list, expressions: List[ExpressionInfo]
    ) -> None:
        for expression in expressions:
            if not expression["isValid"]:
                continue

            name = expression["name"]
            description = None
            if "description" in expression.keys():
                description = expression["description"]

            self._add_expression(vector_data, name, description)

    # pylint: disable=too-many-statements
    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            Output(self.uuid("graph"), "figure"),
            [
                Input(self.uuid("vectors"), "selectedNodes"),
                Input(self.uuid("ensemble"), "value"),
                Input(self.uuid("mode"), "value"),
                Input(self.uuid("base_ens"), "value"),
                Input(self.uuid("delta_ens"), "value"),
                Input(self.uuid("statistics"), "value"),
                Input(self.uuid("cum_interval"), "value"),
                Input(self.uuid("date"), "data"),
                Input(self.uuid("trace_options"), "value"),
                Input(self.uuid("stat_options"), "value"),
                Input(
                    self.uuid("vector_calculator_expressions"),
                    "data",
                ),
            ],
        )
        # pylint: disable=too-many-arguments, too-many-locals, too-many-branches
        def _update_graph(
            vectors: List[str],
            ensembles: List[str],
            calc_mode: str,
            base_ens: str,
            delta_ens: str,
            visualization: str,
            cum_interval: str,
            stored_date: str,
            trace_options: List[str],
            stat_options: List[str],
            expressions: List[ExpressionInfo],
        ) -> dict:
            """Callback to update all graphs based on selections"""

            if not isinstance(ensembles, list):
                raise TypeError("ensembles should always be of type list")

            if calc_mode not in ["ensembles", "delta_ensembles"]:
                raise PreventUpdate

            if vectors is None:
                vectors = self.plot_options.get("vectors", [self.smry_cols[0]])

            # Retreive selected expressions
            selected_expressions = get_selected_expressions(expressions, vectors)
            calculated_units = pd.Series()
            if self.smry_meta is not None:
                calculated_units = get_calculated_units(
                    selected_expressions, self.smry_meta["unit"]
                )

            # Synthesize ensembles list for delta mode
            if calc_mode == "delta_ensembles":
                ensembles = [base_ens, delta_ens]

            # Retrieve previous/current selected date
            date = json.loads(stored_date) if stored_date else None

            # Titles for subplots
            # TODO(Sigurd)
            # Added None to union since date_to_interval_conversion() may return None.
            # Need input on what should be done since a None title is probably not what we want
            titles: List[Union[str, None]] = []
            for vec in vectors:
                if sys.version_info >= (3, 9):
                    unit_vec = vec.removeprefix("AVG_").removeprefix("INTVL_")
                else:
                    unit_vec = (
                        vec[4:]
                        if vec.startswith("AVG_")
                        else (vec[6:] if vec.startswith("INTVL_") else vec)
                    )
                if self.smry_meta is None:
                    titles.append(simulation_vector_description(vec))
                elif vec in calculated_units:
                    titles.append(f"{vec}" f" [{calculated_units[vec]}]")
                else:
                    titles.append(
                        f"{simulation_vector_description(vec)}"
                        f" [{simulation_unit_reformat(self.smry_meta.unit[unit_vec])}]"
                    )
                if "Histogram" in trace_options:
                    titles.append(
                        date_to_interval_conversion(
                            date=date,
                            vector=vec,
                            interval=cum_interval,
                            as_date=False,
                        )
                    )

            # Make a plotly subplot figure
            fig = make_subplots(
                rows=max(1, len(vectors)),
                cols=2 if "Histogram" in trace_options else 1,
                shared_xaxes=True,
                vertical_spacing=0.05,
                subplot_titles=titles if titles else ["No vector selected"],
            )

            # Loop through each vector and calculate relevant plot
            legends = []
            dfs = calculate_vector_dataframes(
                smry=self.smry,
                smry_meta=self.smry_meta,
                ensembles=ensembles,
                vectors=vectors,
                selected_expressions=selected_expressions,
                calc_mode=calc_mode,
                visualization=visualization,
                time_index=self.time_index,
                cum_interval=cum_interval,
            )
            for i, vector in enumerate(vectors):
                if dfs[vector]["data"].empty:
                    continue
                line_shape = get_simulation_line_shape(
                    line_shape_fallback=self.line_shape_fallback,
                    vector=vector,
                    smry_meta=self.smry_meta,
                )
                if visualization == "fanchart":
                    traces = _get_fanchart_traces(
                        dfs[vector]["stat"],
                        vector,
                        colors=self.ens_colors,
                        line_shape=line_shape,
                        interval=cum_interval,
                    )
                elif visualization == "statistics":
                    traces = _add_statistics_traces(
                        dfs[vector]["stat"],
                        vector,
                        colors=self.ens_colors,
                        line_shape=line_shape,
                        interval=cum_interval,
                        stat_options=stat_options,
                    )

                elif visualization == "realizations":
                    traces = add_realization_traces(
                        dfs[vector]["data"],
                        vector,
                        colors=self.ens_colors,
                        line_shape=line_shape,
                        interval=cum_interval,
                    )
                else:
                    raise PreventUpdate

                if "Histogram" in trace_options:
                    histdata = add_histogram_traces(
                        dfs[vector]["data"],
                        vector,
                        date=date,
                        colors=self.ens_colors,
                        interval=cum_interval,
                    )
                    for trace in histdata:
                        fig.add_trace(trace, i + 1, 2)
                if "History" in trace_options:
                    historical_vector_name = historical_vector(
                        vector=vector, smry_meta=self.smry_meta
                    )
                    if (
                        historical_vector_name
                        and historical_vector_name in dfs[vector]["data"].columns
                        and not calc_mode == "delta_ensembles"
                    ):
                        traces.append(
                            add_history_trace(
                                dfs[vector]["data"],
                                historical_vector_name,
                                line_shape,
                            )
                        )

                # Remove unwanted legends(only keep one for each ensemble)
                for trace in traces:
                    if trace.get("showlegend"):
                        if trace.get("legendgroup") in legends:
                            trace["showlegend"] = False
                        else:
                            legends.append(trace.get("legendgroup"))
                    fig.add_trace(trace, i + 1, 1)
                # Add observations
                if calc_mode != "delta_ensembles" and self.observations.get(vector):
                    for trace in add_observation_trace(self.observations.get(vector)):
                        fig.add_trace(trace, i + 1, 1)

            # Remove linked x-axis for histograms
            # Keep uirevision (e.g. zoom) for unchanged data.
            fig.update_xaxes(uirevision="locked")  # Time axis state kept
            for i, vector in enumerate(vectors, start=1):
                if "Histogram" in trace_options:
                    fig.update_xaxes(
                        row=i,
                        col=2,
                        matches=None,
                        showticklabels=True,
                        uirevision=vector,
                    )
                    fig.update_yaxes(row=i, col=2, uirevision=vector)
                fig.update_yaxes(row=i, col=1, uirevision=vector)

            fig = fig.to_dict()
            fig["layout"].update(
                barmode="overlay",
                bargap=0.01,
                bargroupgap=0.2,
            )
            fig["layout"] = self.theme.create_themed_layout(fig["layout"])
            return fig

        @app.callback(
            self.plugin_data_output,
            [self.plugin_data_requested],
            [
                State(self.uuid("vectors"), "selectedNodes"),
                State(self.uuid("ensemble"), "value"),
                State(self.uuid("mode"), "value"),
                State(self.uuid("base_ens"), "value"),
                State(self.uuid("delta_ens"), "value"),
                State(self.uuid("statistics"), "value"),
                State(self.uuid("cum_interval"), "value"),
                State(
                    self.uuid("vector_calculator_expressions"),
                    "data",
                ),
            ],
        )
        def _user_download_data(
            data_requested: Union[int, None],
            vectors: List[str],
            ensembles: List[str],
            calc_mode: str,
            base_ens: str,
            delta_ens: str,
            visualization: str,
            cum_interval: str,
            expressions: List[ExpressionInfo],
        ) -> Union[EncodedFile, str]:

            """Callback to download data based on selections"""
            if data_requested is None:
                raise PreventUpdate

            # Ensure selected ensembles is a list and prevent update if invalid calc_mode
            if calc_mode == "delta_ensembles":
                ensembles = [base_ens, delta_ens]
            elif calc_mode == "ensembles":
                if not isinstance(ensembles, list):
                    raise TypeError("ensembles should always be of type list")
            else:
                raise PreventUpdate

            if vectors is None:
                vectors = self.plot_options.get("vectors", [self.smry_cols[0]])

            # Calculate selected expressions:
            selected_expressions = get_selected_expressions(expressions, vectors)

            dfs = calculate_vector_dataframes(
                smry=self.smry,
                smry_meta=self.smry_meta,
                ensembles=ensembles,
                vectors=vectors,
                selected_expressions=selected_expressions,
                calc_mode=calc_mode,
                visualization=visualization,
                time_index=self.time_index,
                cum_interval=cum_interval,
            )
            for vector, df in dfs.items():
                if visualization in ["fanchart", "statistics"]:
                    df["stat"] = df["stat"].sort_values(
                        by=[("", "ENSEMBLE"), ("", "DATE")]
                    )
                    if vector.startswith(("AVG_", "INTVL_")):
                        df["stat"]["", "DATE"] = df["stat"]["", "DATE"].astype(str)
                        df["stat"]["", "DATE"] = df["stat"]["", "DATE"].apply(
                            date_to_interval_conversion,
                            vector=vector,
                            interval=cum_interval,
                            as_date=False,
                        )
                else:
                    df["data"] = df["data"].sort_values(by=["ENSEMBLE", "REAL", "DATE"])
                    # Reorder columns
                    df["data"] = df["data"][
                        ["ENSEMBLE", "REAL", "DATE"]
                        + [
                            col
                            for col in df["data"].columns
                            if col not in ["ENSEMBLE", "REAL", "DATE"]
                        ]
                    ]
                    if vector.startswith(("AVG_", "INTVL_")):
                        df["data"]["DATE"] = df["data"]["DATE"].astype(str)
                        df["data"]["DATE"] = df["data"]["DATE"].apply(
                            date_to_interval_conversion,
                            vector=vector,
                            interval=cum_interval,
                            as_date=False,
                        )

            # : is replaced with _ in filenames to stay within POSIX portable pathnames
            # (e.g. : is not valid in a Windows path)
            return WebvizPluginABC.plugin_data_compress(
                [
                    {
                        "filename": f"{vector.replace(':', '_')}.csv",
                        "content": df.get("stat", df["data"]).to_csv(index=False),
                    }
                    for vector, df in dfs.items()
                ]
            )

        @app.callback(
            [
                Output(self.uuid("show_ensembles"), "style"),
                Output(self.uuid("calc_delta"), "style"),
            ],
            [Input(self.uuid("mode"), "value")],
        )
        def _update_mode(mode: str) -> Tuple[dict, dict]:
            """Switch displayed ensemble selector for delta/no-delta"""
            if mode == "ensembles":
                style = {"display": "block"}, {"display": "none"}
            else:
                style = {"display": "none"}, {"display": "block"}
            return style

        @app.callback(
            Output(self.uuid("view_stat_options"), "style"),
            [Input(self.uuid("statistics"), "value")],
        )
        def _update_view_stat_options(visualization: str) -> dict:
            """Only show statistics picker if in statistics mode"""
            return (
                {"display": "block"}
                if visualization == "statistics"
                else {"display": "none"}
            )

        @app.callback(
            Output(self.uuid("date"), "data"),
            [Input(self.uuid("graph"), "clickData")],
            [State(self.uuid("date"), "data")],
        )
        def _update_date(clickdata: dict, date: str) -> str:
            """Store clicked date for use in other callback"""
            date = clickdata["points"][0]["x"] if clickdata else json.loads(date)
            return json.dumps(date)

        @app.callback(
            Output(self.uuid("cum_interval"), "options"),
            [
                Input(self.uuid("vectors"), "selectedNodes"),
            ],
            [State(self.uuid("cum_interval"), "options")],
        )
        def _activate_interval_radio_buttons(
            vectors: List[str],
            options: List[dict],
        ) -> List[dict]:
            """Switch activate/deactivate radio buttons for selectibg interval for
            calculations from cumulatives"""
            active = False
            if vectors:
                for vector in vectors:
                    if vector is not None and vector.startswith(("AVG_", "INTVL_")):
                        active = True
                        break
            if active:
                return [dict(option, **{"disabled": False}) for option in options]
            return [dict(option, **{"disabled": True}) for option in options]

        @app.callback(
            Output(self.uuid("vector_calculator"), "externalParseData"),
            Input(self.uuid("vector_calculator"), "externalParseExpression"),
        )
        def _parse_vector_calculator_expression(
            expression: ExpressionInfo,
        ) -> ExternalParseData:
            if expression is None:
                raise PreventUpdate

            return wsc.VectorCalculator.external_parse_data(expression)

        @app.callback(
            [
                Output(self.uuid("vector_calculator_expressions"), "data"),
                Output(self.uuid("vectors"), "data"),
                Output(self.uuid("vectors"), "selectedTags"),
                Output(self.uuid("vectors"), "customVectorDefinitions"),
            ],
            Input(self.uuid("modal_vector_calculator"), "is_open"),
            [
                State(self.uuid("vector_calculator_expressions_modal_open"), "data"),
                State(self.uuid("vector_calculator_expressions"), "data"),
                State(self.uuid("vectors"), "selectedNodes"),
                State(self.uuid("vectors"), "customVectorDefinitions"),
            ],
        )
        def _update_vector_calculator_expressions_actual(
            modal_open: bool,
            new_expressions: List[ExpressionInfo],
            existing_expressions: List[ExpressionInfo],
            selected_vectors: List[str],
            custom_vector_definitions: dict,
        ) -> list:
            if modal_open or (new_expressions == existing_expressions):
                raise PreventUpdate

            # Deep copy to prevent modifying self.vector_data
            vector_data = copy.deepcopy(self.vector_data)
            self._add_expressions_to_vector_data(vector_data, new_expressions)

            new_selected_vectors = self._get_valid_vector_selections(
                vector_data, selected_vectors, new_expressions, existing_expressions
            )

            # Prevent updates if selected vectors are unchanged
            if new_selected_vectors == selected_vectors:
                new_selected_vectors = dash.no_update

            new_custom_vector_definitions = (
                get_custom_vector_definitions_from_expressions(new_expressions)
            )

            if new_custom_vector_definitions == custom_vector_definitions:
                new_custom_vector_definitions = dash.no_update

            return [
                new_expressions,
                vector_data,
                new_selected_vectors,
                new_custom_vector_definitions,
            ]

        @app.callback(
            Output(self.uuid("vector_calculator_expressions_modal_open"), "data"),
            Input(self.uuid("vector_calculator"), "expressions"),
        )
        def _update_vector_calculator_expressions(
            expressions: List[ExpressionInfo],
        ) -> list:
            valid_expressions: List[ExpressionInfo] = [
                elm for elm in expressions if elm["isValid"]
            ]

            return valid_expressions

        @app.callback(
            Output(self.uuid("modal_vector_calculator"), "is_open"),
            [
                Input(self.uuid("vector_calculator_open_btn"), "n_clicks"),
            ],
            [State(self.uuid("modal_vector_calculator"), "is_open")],
        )
        def _toggle_modal(n_open_clicks: int, is_open: bool) -> Optional[bool]:
            if n_open_clicks:
                return not is_open
            raise PreventUpdate

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        functions: List[Tuple[Callable, list]] = []
        if self.csvfile:
            functions.append((read_csv, [{"csv_file": self.csvfile}]))
        else:
            functions.extend(self.emodel.webvizstore)
        if self.obsfile:
            functions.append((get_path, [{"path": self.obsfile}]))
        if self.predefined_expressions_path:
            functions.append((get_path, [{"path": self.predefined_expressions_path}]))
        return functions


# pylint: disable = too-many-arguments
@CACHE.memoize(timeout=CACHE.TIMEOUT)
def calculate_vector_dataframes(
    smry: pd.DataFrame,
    smry_meta: Union[pd.DataFrame, None],
    ensembles: List[str],
    vectors: List[str],
    selected_expressions: List[ExpressionInfo],
    calc_mode: str,
    visualization: str,
    time_index: str,
    cum_interval: str,
) -> dict:
    """Wraps cached function for individual vectors"""
    return {
        vector: calculate_vector_dataframe(
            smry=smry,
            smry_meta=smry_meta,
            ensembles=ensembles,
            vector=vector,
            selected_expressions=selected_expressions,
            calc_mode=calc_mode,
            visualization=visualization,
            time_index=time_index,
            cum_interval=cum_interval,
        )
        for vector in vectors
    }


# pylint: disable = too-many-arguments
@CACHE.memoize(timeout=CACHE.TIMEOUT)
def calculate_vector_dataframe(
    smry: pd.DataFrame,
    smry_meta: Union[pd.DataFrame, None],
    ensembles: List[str],
    vector: str,
    selected_expressions: List[ExpressionInfo],
    calc_mode: str,
    visualization: str,
    time_index: str,
    cum_interval: str,
) -> Dict[str, pd.DataFrame]:
    expression = get_expression_from_name(vector, selected_expressions)
    if expression:
        data = get_calculated_vector_df(expression, smry, ensembles)
    elif vector.startswith("AVG_"):
        total_vector = f"{vector[4:7] + vector[7:].replace('R', 'T', 1)}"
        data = filter_df(smry, ensembles, total_vector, smry_meta, calc_mode)
        data = calc_from_cumulatives(
            data=data,
            column_keys=total_vector,
            time_index=cum_interval,
            time_index_input=time_index,
            as_rate=True,
        )
        vector = rename_vec_from_cum(vector=vector[4:], as_rate=True)
    elif vector.startswith("INTVL_"):
        total_vector = vector.lstrip("INTVL_")
        data = filter_df(smry, ensembles, total_vector, smry_meta, calc_mode)
        data = calc_from_cumulatives(
            data=data,
            column_keys=total_vector,
            time_index=cum_interval,
            time_index_input=time_index,
            as_rate=False,
        )
    else:
        data = filter_df(smry, ensembles, vector, smry_meta, calc_mode)

    if calc_mode == "delta_ensembles":
        data = calculate_delta(data, ensembles[0], ensembles[1])

    output: Dict[str, pd.DataFrame] = {"data": data}
    if visualization in [
        "statistics",
        "statistics_hist",
        "fanchart",
        "fanchart_hist",
    ]:
        output["stat"] = calc_series_statistics(data, [vector])
    return output


def filter_df(
    df: pd.DataFrame,
    ensembles: List[str],
    vector: str,
    smry_meta: Union[pd.DataFrame, None],
    calc_mode: str,
) -> pd.DataFrame:
    """Filter dataframe for current vector. Include history
    vector if present"""
    columns = ["REAL", "ENSEMBLE", "DATE", vector]
    historical_vector_name = historical_vector(vector=vector, smry_meta=smry_meta)
    if (
        historical_vector_name
        and historical_vector_name in df.columns
        and not "delta" in calc_mode
    ):
        columns.append(historical_vector_name)
    fdf = df[columns].copy()
    if historical_vector_name in fdf.columns:
        fdf.loc[
            fdf["DATE"].astype("str") > str(datetime.datetime.now()),
            historical_vector_name,
        ] = np.nan
    return fdf.loc[fdf["ENSEMBLE"].isin(ensembles)]


def calculate_delta(df: pd.DataFrame, base_ens: str, delta_ens: str) -> pd.DataFrame:
    """Calculate delta between two ensembles"""
    base_df = (
        df.loc[df["ENSEMBLE"] == base_ens]
        .set_index(["DATE", "REAL"])
        .drop("ENSEMBLE", axis=1)
    )
    delta_df = (
        df.loc[df["ENSEMBLE"] == delta_ens]
        .set_index(["DATE", "REAL"])
        .drop("ENSEMBLE", axis=1)
    )
    dframe = base_df.sub(delta_df).reset_index()
    dframe["ENSEMBLE"] = f"({base_ens}) - ({delta_ens})"
    return dframe.dropna(axis=0, how="any")


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def add_histogram_traces(
    dframe: pd.DataFrame,
    vector: str,
    date: Union[str, None],
    colors: dict,
    interval: str,
) -> List[dict]:
    """Renders a histogram trace per ensemble for a given date"""
    dframe[("DATE")] = dframe[("DATE")].astype(str)
    date = date_to_interval_conversion(
        date=date, vector=vector, interval=interval, as_date=True
    )
    data = dframe.loc[dframe[("DATE")] == date]

    return [
        {
            "type": "histogram",
            "x": list(ens_df[vector]),
            "name": ensemble,
            "marker": {
                "color": colors.get(ensemble, colors[list(colors.keys())[0]]),
                "line": {"color": "black", "width": 1},
            },
            "showlegend": False,
            "legendgroup": ensemble,
        }
        for ensemble, ens_df in data.groupby(("ENSEMBLE"))
    ]


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def add_observation_trace(obs: dict) -> List[dict]:
    return [
        {
            "x": [value.get("date"), []],
            "y": [value.get("value"), []],
            "marker": {"color": "black"},
            "text": value.get("comment", None),
            "hoverinfo": "y+x+text",
            "showlegend": False,
            "error_y": {
                "type": "data",
                "array": [value.get("error"), []],
                "visible": True,
            },
        }
        for value in obs.get("observations", [])
    ]


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def add_realization_traces(
    dframe: pd.DataFrame, vector: str, colors: dict, line_shape: str, interval: str
) -> List[dict]:
    """Renders line trace for each realization, includes history line if present"""
    hovertemplate = render_hovertemplate(vector, interval)
    return [
        {
            "line": {"shape": line_shape},
            "x": list(real_df["DATE"]),
            "y": list(real_df[vector]),
            "hovertemplate": f"{hovertemplate}Realization: {real}, Ensemble: {ensemble}",
            "name": ensemble,
            "legendgroup": ensemble,
            "marker": {"color": colors.get(ensemble, colors[list(colors.keys())[0]])},
            "showlegend": real_no == 0,
        }
        for ens_no, (ensemble, ens_df) in enumerate(dframe.groupby("ENSEMBLE"))
        for real_no, (real, real_df) in enumerate(ens_df.groupby("REAL"))
    ]


def add_history_trace(dframe: pd.DataFrame, vector: str, line_shape: str) -> dict:
    """Renders the history line"""
    df = dframe.loc[
        (dframe["REAL"] == dframe["REAL"].unique()[0])
        & (dframe["ENSEMBLE"] == dframe["ENSEMBLE"].unique()[0])
    ]
    return {
        "line": {"shape": line_shape},
        "x": df["DATE"],
        "y": df[vector],
        "hovertext": "History",
        "hoverinfo": "y+x+text",
        "name": "History",
        "marker": {"color": "black"},
        "showlegend": True,
        "legendgroup": "History",
    }


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def _get_fanchart_traces(
    stat_df: pd.DataFrame, vector: str, colors: dict, line_shape: str, interval: str
) -> list:
    """Add fanchart traces for selected vector"""
    traces = []
    for ensemble, ens_df in stat_df.groupby(("", "ENSEMBLE")):

        data = FanchartData(
            samples=ens_df[("", "DATE")].tolist(),
            low_high=LowHighData(
                low_data=ens_df[(vector, "low_p90")].values,
                low_name="P90",
                high_data=ens_df[(vector, "high_p10")].values,
                high_name="P10",
            ),
            minimum_maximum=MinMaxData(
                minimum=ens_df[(vector, "min")].values,
                maximum=ens_df[(vector, "max")].values,
            ),
            free_line=FreeLineData("Mean", ens_df[(vector, "mean")].values),
        )
        traces.extend(
            get_fanchart_traces(
                data=data,
                color=colors.get(ensemble, colors[list(colors.keys())[0]]),
                legend_group=ensemble,
                line_shape=line_shape,
                hovertemplate=render_hovertemplate(vector, interval),
            )
        )
    return traces


def _add_statistics_traces(
    stat_df: pd.DataFrame,
    vector: str,
    colors: dict,
    line_shape: str,
    interval: str,
    stat_options: List[str],
) -> list:
    """Add fanchart traces for selected vector"""
    traces = []
    for ensemble, ens_df in stat_df.groupby(("", "ENSEMBLE")):
        traces.extend(
            add_statistics_traces(
                ens_stat_df=ens_df,
                vector=vector,
                color=colors.get(ensemble, colors[list(colors.keys())[0]]),
                legend_group=ensemble,
                line_shape=line_shape,
                hovertemplate=render_hovertemplate(vector=vector, interval=interval),
                stat_options=stat_options,
            )
        )
    return traces


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=False)


@webvizstore
def get_path(path: Path) -> Path:
    return Path(path)



from dash import html
import webviz_core_components as wcc

from webviz_subsurface._models import InplaceVolumesModel


def distributions_main_layout(uuid: str, volumemodel: InplaceVolumesModel) -> html.Div:
    return html.Div(
        children=[
            html.Div(
                id={"id": uuid, "page": "custom"},
                style={"display": "none"},
                children=custom_plotting_layout(uuid),
            ),
            html.Div(
                id={"id": uuid, "page": "1p1t"},
                style={"display": "block"},
                children=one_plot_one_table_layout(uuid),
            ),
            html.Div(
                id={"id": uuid, "page": "per_zr"},
                style={"display": "none"},
                children=plots_per_zone_region_layout(uuid, volumemodel),
            ),
            html.Div(
                id={"id": uuid, "page": "conv"},
                style={"display": "none"},
                children=convergence_plot_layout(uuid),
            ),
        ]
    )


def table_main_layout(uuid: str) -> html.Div:
    return wcc.Frame(
        id={"id": uuid, "wrapper": "table", "page": "table"},
        color="white",
        highlight=False,
        style={"height": "91vh"},
        children=[],
    )


def convergence_plot_layout(uuid: str) -> html.Div:
    return wcc.Frame(
        color="white",
        highlight=False,
        style={"height": "91vh"},
        children=[
            wcc.Graph(
                id={"id": uuid, "element": "plot", "page": "conv"},
                config={"displayModeBar": False},
                style={"height": "85vh"},
            )
        ],
    )


def custom_plotting_layout(uuid: str) -> html.Div:
    return wcc.Frame(
        color="white",
        highlight=False,
        style={"height": "91vh"},
        children=[
            wcc.RadioItems(
                id={"id": uuid, "element": "plot-table-select"},
                options=[
                    {"label": "Plot", "value": "graph"},
                    {"label": "Table", "value": "table"},
                ],
                value="graph",
                vertical=False,
            ),
            html.Div(
                id={"id": uuid, "wrapper": "graph", "page": "custom"},
                style={"display": "inline"},
                children=wcc.Graph(
                    id={"id": uuid, "element": "graph", "page": "custom"},
                    config={"displayModeBar": False},
                    style={"height": "85vh"},
                ),
            ),
            html.Div(
                id={"id": uuid, "wrapper": "table", "page": "custom"},
                style={"display": "none"},
            ),
        ],
    )


def one_plot_one_table_layout(uuid: str) -> html.Div:
    return html.Div(
        children=[
            wcc.Frame(
                color="white",
                highlight=False,
                style={"height": "44vh"},
                children=wcc.Graph(
                    id={"id": uuid, "element": "graph", "page": "1p1t"},
                    config={"displayModeBar": False},
                    style={"height": "44vh"},
                ),
            ),
            wcc.Frame(
                color="white",
                highlight=False,
                style={"height": "44vh"},
                id={"id": uuid, "wrapper": "table", "page": "1p1t"},
                children=[],
            ),
        ]
    )


def plots_per_zone_region_layout(
    uuid: str, volumemodel: InplaceVolumesModel
) -> html.Div:
    selectors = [
        x
        for x in ["ZONE", "REGION", "FACIES", "FIPNUM", "SET"]
        if x in volumemodel.selectors
    ]
    height = "42vh" if len(selectors) < 3 else "25vh"
    layout = []
    for selector in selectors:
        layout.append(
            wcc.Frame(
                color="white",
                highlight=False,
                style={"height": height},
                children=wcc.FlexBox(
                    children=[
                        html.Div(
                            style={"flex": 1},
                            children=wcc.Graph(
                                id={
                                    "id": uuid,
                                    "chart": "pie",
                                    "selector": selector,
                                    "page": "per_zr",
                                },
                                config={"displayModeBar": False},
                                style={"height": height},
                            ),
                        ),
                        html.Div(
                            style={"flex": 3},
                            children=wcc.Graph(
                                id={
                                    "id": uuid,
                                    "chart": "bar",
                                    "selector": selector,
                                    "page": "per_zr",
                                },
                                config={"displayModeBar": False},
                                style={"height": height},
                            ),
                        ),
                    ]
                ),
            )
        )

    return html.Div(layout)



from typing import Dict, Callable, Optional, List

from dash import html, dcc
import webviz_core_components as wcc

from .property_qc_view import property_qc_view
from .property_delta_view import property_delta_view
from .property_response_view import property_response_view

from ..models import PropertyStatisticsModel


def main_view(
    get_uuid: Callable,
    property_model: PropertyStatisticsModel,
    vector_options: List[Dict],
    surface_folders: Optional[Dict] = None,
) -> dcc.Tabs:
    tabs = [
        make_tab(
            label="Property QC",
            children=property_qc_view(get_uuid=get_uuid, property_model=property_model),
        )
    ]
    if len(property_model.ensembles) > 1:
        tabs.append(
            make_tab(
                label="AHM impact on property",
                children=property_delta_view(
                    get_uuid=get_uuid,
                    property_model=property_model,
                    surface_folders=surface_folders,
                ),
            )
        )
    tabs.append(
        make_tab(
            label="Property impact on simulation profiles",
            children=property_response_view(
                get_uuid=get_uuid,
                property_model=property_model,
                vector_options=vector_options,
                surface_folders=surface_folders,
            ),
        ),
    )

    return html.Div(
        id=get_uuid("layout"),
        children=wcc.Tabs(
            style={"width": "100%"},
            children=tabs,
        ),
    )


def make_tab(label: str, children: wcc.FlexBox) -> dcc.Tab:

    return wcc.Tab(
        label=label,
        children=children,
    )



#!/usr/bin/env python
"""Convert UNSMRY file to Apache Arrow IPC file format (also known as Feather V2)
"""

from typing import List, Iterable, Dict
import argparse
import re
from pathlib import Path
import logging
import json
import time

from ecl.summary import EclSum, EclSumKeyWordVector
import pyarrow as pa
from pyarrow import feather

logger = logging.getLogger(__name__)


DESCRIPTION: str = """
Convert UNSMRY file to Apache Arrow IPC file format (also known as Feather V2)
"""
CATEGORY: str = "utility.eclipse"
EXAMPLES: str = """
Convert and output summary data in SOME.UNSMRY file to an Arrow IPC file named unsmry.arrow
this in the ert workflow:

    FORWARD_MODEL SMRY2ARROW(<INPUT>=eclipse/model/SOME.UNSMRY)

"""


def _get_parser() -> argparse.ArgumentParser:
    """Setup parser for command line options"""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=DESCRIPTION,
    )
    parser.add_argument(
        "eclbase",
        type=Path,
        help="Eclipse base name",
    )
    parser.add_argument(
        "--output",
        type=Path,
        help="Output file",
        default=Path() / "share" / "results" / "tables" / "unsmry.arrow",
    )
    return parser


# This logic is copied from well_connection_status.py.
# Should be refactored so that the two modules share a common implementation.
def _is_cpi_column(column_name: str) -> bool:
    return bool(re.match("^CPI:[A-Z0-9_-]{1,8}:[0-9]+,[0-9]+,[0-9]+$", column_name))


def _create_smry_meta_dict(
    eclsum: EclSum, column_names: Iterable[str]
) -> Dict[str, dict]:
    """Builds dictionary containing metadata for all the specified summary columns"""
    smry_meta = {}

    for col_name in column_names:
        col_meta = {}
        col_meta["unit"] = eclsum.unit(col_name)
        col_meta["is_total"] = eclsum.is_total(col_name)
        col_meta["is_rate"] = eclsum.is_rate(col_name)
        col_meta["is_historical"] = eclsum.smspec_node(col_name).is_historical()
        col_meta["keyword"] = eclsum.smspec_node(col_name).keyword
        col_meta["wgname"] = eclsum.smspec_node(col_name).wgname

        num = eclsum.smspec_node(col_name).get_num()
        if num is not None:
            col_meta["get_num"] = num

        smry_meta[col_name] = col_meta

    return smry_meta


def _load_smry_into_table(smry_filename: str) -> pa.Table:
    """
    Reads data from SMRY file into PyArrow Table.
    DATE column is stored as an Arrow timetamp with ms resolution, timestamp[ms]
    All numeric columns will be stored as 32 bit float
    Summary meta data will be attached per field/column of the table's schema under the
    'smry_meta' key
    """

    eclsum = EclSum(smry_filename, include_restart=False, lazy_load=False)

    # For now, we go via a set to prune out duplicate entries being returned by EclSumKeyWordVector,
    # see: https://github.com/equinor/ecl/issues/816#issuecomment-865881283
    column_names: List[str] = list(set(EclSumKeyWordVector(eclsum, add_keywords=True)))

    # Exclude CPI columns from export
    org_col_count = len(column_names)
    column_names = [colname for colname in column_names if not _is_cpi_column(colname)]
    if len(column_names) != org_col_count:
        logger.info(
            f"Excluding {org_col_count - len(column_names)} CPI columns from export"
        )

    # Fetch the dates as a numpy array with ms resolution
    np_dates_ms = eclsum.numpy_dates

    smry_meta_dict = _create_smry_meta_dict(eclsum, column_names)

    # Datatypes to use for DATE column and all the numeric columns
    dt_timestamp_ms = pa.timestamp("ms")
    dt_float32 = pa.float32()

    # Build schema for the table
    field_list: List[pa.Field] = []
    field_list.append(pa.field("DATE", dt_timestamp_ms))
    for colname in column_names:
        field_metadata = {b"smry_meta": json.dumps(smry_meta_dict[colname])}
        field_list.append(pa.field(colname, dt_float32, metadata=field_metadata))

    schema = pa.schema(field_list)

    # Now extract all the summary vectors one by one
    # We do this through EclSum.numpy_vector() instead of EclSum.pandas_frame() since
    # the latter throws an exception if the SMRY data has timestamps beyond 2262,
    # see: https://github.com/equinor/ecl/issues/802
    column_arrays = [np_dates_ms]

    for colname in column_names:
        colvector = eclsum.numpy_vector(colname)
        column_arrays.append(colvector)

    table = pa.table(column_arrays, schema=schema)

    return table


def smry2arrow(smry_filename: Path, arrow_filename: Path) -> None:
    start_s = time.perf_counter()

    logger.debug(f"Reading SMRY data from: {smry_filename}")
    table: pa.Table = _load_smry_into_table(str(smry_filename))
    read_s = time.perf_counter() - start_s

    # Writing here is done through the feather import, but could also be done using
    # pa.RecordBatchFileWriter.write_table() with a few pa.ipc.IpcWriteOptions().
    # It is convenient to use feather since it has ready configured defaults and the
    # actual file format is the same (https://arrow.apache.org/docs/python/feather.html)
    logger.debug(f"Writing {table.shape} arrow file to: {arrow_filename}")
    feather.write_feather(table, dest=arrow_filename)

    logger.debug(
        f"Conversion took: {(time.perf_counter() - start_s):.2f}s (read={read_s:.2f}s)"
    )


def main() -> None:
    """Entry point from command line"""
    parser = _get_parser()
    args = parser.parse_args()

    # Create the output folder if it doesn't exist
    args.output.parent.mkdir(parents=True, exist_ok=True)

    smry2arrow(args.eclbase.with_suffix(".UNSMRY"), args.output)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    main()



from typing import Dict, List
from pathlib import Path
import json

from dash import html, Dash, callback_context, Input, Output, ALL
from webviz_config import WebvizPluginABC, WebvizSettings
import webviz_core_components as wcc

from webviz_subsurface._providers import EnsembleTableProviderFactory
from webviz_subsurface._components.tornado.tornado_widget import TornadoWidget
from webviz_subsurface._datainput.fmu_input import find_sens_type


class TornadoPlotterFMU(WebvizPluginABC):
    """General tornado plotter for FMU data from a csv file of responses.
    ---
    * **`ensembles`:** Which ensembles in `shared_settings` to visualize.
    * **`csvfile`:** Relative ensemble path to csv file with responses
    * **`aggregated_csvfile`:** Alternative to ensemble + csvfile with
    aggregated responses. Requires REAL and ENSEMBLE columns
    * **`aggregated_parameterfile`:** Necessary when aggregated_csvfile
    is specified. File with sensitivity specification for each realization.
    Requires columns REAL, ENSEMBLE, SENSNAME and SENSCASE.
    * **`initial_response`:** Initialize plugin with this response column
    visualized
    * **`single_value_selectors`:** List of columns in response csv file
    that should be used to select/filter data. E.g. for UNSMRY data the DATE
    column can be used. For each entry a Dropdown is shown with all unique
    values and a single value can be selected at a time.
    * **`multi_value_selectors`:** List of columns in response csv file
    to filter/select data. For each entry a Select is shown with
    all unique values. Multiple values can be selected at a time,
    and a tornado plot will be shown from the matching response rows.
    Used e.g. for volumetrics data, to select a subset of ZONES and
    REGIONS.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        csvfile: str = None,
        ensemble: str = None,
        aggregated_csvfile: Path = None,
        aggregated_parameterfile: Path = None,
        initial_response: str = None,
        single_value_selectors: List[str] = None,
        multi_value_selectors: List[str] = None,
    ):
        super().__init__()
        self._single_filters = single_value_selectors if single_value_selectors else []
        self._multi_filters = multi_value_selectors if multi_value_selectors else []
        provider = EnsembleTableProviderFactory.instance()

        if ensemble is not None and csvfile is not None:
            ensemble_dict: Dict[str, str] = {
                ensemble: webviz_settings.shared_settings["scratch_ensembles"][ensemble]
            }
            self._parameterproviderset = (
                provider.create_provider_set_from_per_realization_parameter_file(
                    ensemble_dict
                )
            )
            self._tableproviderset = (
                provider.create_provider_set_from_per_realization_csv_file(
                    ensemble_dict, csvfile
                )
            )
            self._ensemble_name = ensemble
        elif aggregated_csvfile and aggregated_parameterfile is not None:
            self._tableproviderset = (
                provider.create_provider_set_from_aggregated_csv_file(
                    aggregated_csvfile
                )
            )
            self._parameterproviderset = (
                provider.create_provider_set_from_aggregated_csv_file(
                    aggregated_parameterfile
                )
            )
            if len(self._tableproviderset.ensemble_names()) != 1:
                raise ValueError(
                    "Csv file has multiple ensembles. "
                    "This plugin only supports a single ensemble"
                )
            self._ensemble_name = self._tableproviderset.ensemble_names()[0]
        else:
            raise ValueError(
                "Specify either ensembles and csvfile or aggregated_csvfile "
                "and aggregated_parameterfile"
            )

        try:
            design_matrix_df = self._parameterproviderset.ensemble_provider(
                self._ensemble_name
            ).get_column_data(column_names=["SENSNAME", "SENSCASE"])
        except KeyError as exc:
            raise KeyError(
                "Required columns 'SENSNAME' and 'SENSCASE' is missing "
                f"from {self._ensemble_name}. Cannot calculate tornado plots"
            ) from exc
        design_matrix_df["ENSEMBLE"] = self._ensemble_name
        design_matrix_df["SENSTYPE"] = design_matrix_df.apply(
            lambda row: find_sens_type(row.SENSCASE), axis=1
        )
        self._tornado_widget = TornadoWidget(
            realizations=design_matrix_df, app=app, webviz_settings=webviz_settings
        )
        self._responses: List[str] = self._tableproviderset.ensemble_provider(
            self._ensemble_name
        ).column_names()
        if self._single_filters:
            self._responses = [
                response
                for response in self._responses
                if response not in self._single_filters
            ]
        if self._multi_filters:
            self._responses = [
                response
                for response in self._responses
                if response not in self._multi_filters
            ]
        self._initial_response: str = (
            initial_response if initial_response else self._responses[0]
        )
        self.set_callbacks(app)

    @property
    def single_filter_layout(self) -> html.Div:
        """Creates dropdowns for any data columns added as single value filters"""
        elements = []
        for selector in self._single_filters:
            values = (
                self._tableproviderset.ensemble_provider(self._ensemble_name)
                .get_column_data([selector])[selector]
                .unique()
            )
            elements.append(
                wcc.Dropdown(
                    label=selector,
                    id={
                        "id": self.uuid("selectors"),
                        "name": selector,
                        "type": "single_filter",
                    },
                    options=[{"label": val, "value": val} for val in values],
                    value=values[0],
                    clearable=False,
                )
            )
        return html.Div(children=elements)

    @property
    def multi_filter_layout(self) -> html.Div:
        """Creates wcc.Selects for any data columns added as multi value filters"""
        elements = []
        for selector in self._multi_filters:
            values = (
                self._tableproviderset.ensemble_provider(self._ensemble_name)
                .get_column_data([selector])[selector]
                .unique()
            )
            elements.append(
                wcc.SelectWithLabel(
                    label=selector,
                    id={
                        "id": self.uuid("selectors"),
                        "name": selector,
                        "type": "multi_filter",
                    },
                    options=[{"label": val, "value": val} for val in values],
                    value=values,
                    size=min(5, len(values)),
                )
            )
        return html.Div(children=elements)

    @property
    def response_layout(self) -> html.Div:
        """Creates a labelled dropdown with all response columns"""
        return wcc.Dropdown(
            label="Response",
            id=self.uuid("response"),
            options=[
                {"label": response, "value": response} for response in self._responses
            ],
            value=self._initial_response,
            clearable=False,
        )

    @property
    def layout(self) -> html.Div:
        return wcc.FlexBox(
            children=[
                wcc.Frame(
                    style={"flex": 1},
                    children=[
                        wcc.Selectors(label="Selectors", children=self.response_layout),
                        wcc.Selectors(
                            label="Filters",
                            children=[
                                self.single_filter_layout,
                                self.multi_filter_layout,
                            ],
                        ),
                    ],
                ),
                wcc.Frame(
                    color="white",
                    highlight=False,
                    style={"flex": 5},
                    children=self._tornado_widget.layout,
                ),
            ]
        )

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            Output(self._tornado_widget.storage_id, "data"),
            Input(self.uuid("response"), "value"),
            Input(
                {"id": self.uuid("selectors"), "name": ALL, "type": "single_filter"},
                "value",
            ),
            Input(
                {"id": self.uuid("selectors"), "name": ALL, "type": "multi_filter"},
                "value",
            ),
        )
        def _update_tornado_with_response_values(
            response: str, single_filters: List, multi_filters: List
        ) -> str:
            """Returns a json dump for the tornado plot with the response values per realization"""

            data = self._tableproviderset.ensemble_provider(
                self._ensemble_name
            ).get_column_data([response] + self._single_filters + self._multi_filters)

            # Filter data
            if single_filters is not None:
                for value, input_dict in zip(
                    single_filters, callback_context.inputs_list[1]
                ):
                    data = data.loc[data[input_dict["id"]["name"]] == value]
            if multi_filters is not None:
                for value, input_dict in zip(
                    multi_filters, callback_context.inputs_list[2]
                ):
                    data = data.loc[data[input_dict["id"]["name"]].isin(value)]

            return json.dumps(
                {
                    "ENSEMBLE": self._ensemble_name,
                    "data": data.groupby("REAL")
                    .sum()
                    .reset_index()[["REAL", response]]
                    .values.tolist(),
                    "number_format": "#.4g",
                }
            )
